import kaa_testutils
from kaa import highlight
from kaa.filetype.python import pythonmode


class TestPythonMode(kaa_testutils._TestDocBase):
    DEFAULTMODECLASS = pythonmode.PythonMode

    def test_get_headers(self):
        script = '''
class spam:
    def ham(self, \, x=(def xxx)):
        'def'
# class xyz
def egg():
   """\\\\"""
def bacon():
'''
        doc = self._getdoc(script)
        tokens = [t for t in doc.mode.get_headers()]
        assert tokens == [
            ('namespace', None, 'spam', 'spam', None, 1),
            ('function', tokens[0], 'spam.ham', 'ham()', None, 13),
            ('function', None, 'egg', 'egg()', None, 74),
            ('function', None, 'bacon', 'bacon()', None, 97)]



from kaa.ui.statusbar import statusbarmode
import kaa_testutils


class TestStatusBarMode(kaa_testutils._TestDocBase):

    def _getmodeclass(self):
        return statusbarmode.StatusBarMode

    def test_sbarinfo(self):
        sbarinfo = statusbarmode.StatusInfo()
        sbarinfo.set_info(filename='filename',
                          modified=1)

        doc = self._getdoc('')
        doc.mode.build_status(sbarinfo)

        assert doc.gettext(0, doc.endpos()) == 'filename* [1:1] 1'



from kaa import screen, document
import kaa_testutils


class TestTranslateChars:

    def test_transform(self):
        chrs, cols, positions, intervals = screen.translate_chars(
            10, 'abcdefg', 8, 2)

        assert chrs == 'abcdefg'
        assert cols == [1] * 7
        assert positions == list(range(10, 17))
        assert intervals == [0] * 7

    def test_tab(self):
        chrs, cols, positions, intervals = screen.translate_chars(
            10, 'ab\t\tc', 4, 2)

        assert chrs == 'ab      c'
        assert cols == [1] * 9
        assert positions == [10, 11, 12, 12, 13, 13, 13, 13, 14]
        assert intervals == [0, 0, 0, 1, 0, 1, 2, 3, 0]

    def test_ctrlchars(self):
        chrs, cols, positions, intervals = screen.translate_chars(
            10, '\x01\x02\x03\x7f', 8, 2)

        assert chrs == r'\x01\x02\x03\x7f'
        assert cols == [1] * 16
        assert positions == [
            10, 10, 10, 10, 11, 11, 11, 11,
            12, 12, 12, 12, 13, 13, 13, 13]
        assert intervals == [0, 1, 2, 3] * 4

        chrs, cols, positions, intervals = screen.translate_chars(
            10, '\r\n', 8, 2)

        assert chrs == '\\r\n'
        assert cols == [1] * 3
        assert positions == [10, 10, 11]
        assert intervals == [0, 1, 0]


class TestColumnSplitter:

    def test_split(self):
        rows = screen.col_splitter(
            6, 0, '0123456789', [1] * 10, list(range(10)), [0] * 10, [0] * 10, {})

        row1, row2 = rows
        assert row1.chars == '01234'
        assert row1.cols == [1, 1, 1, 1, 1]
        assert row1.positions == [0, 1, 2, 3, 4]
        assert row1.posfrom == 0
        assert row1.posto == 5

        assert row2.chars == '56789'
        assert row2.cols == [1, 1, 1, 1, 1]
        assert row2.positions == [5, 6, 7, 8, 9]
        assert row2.posfrom == 5
        assert row2.posto == 10

    def test_empty(self):
        rows = screen.col_splitter(5, 0, [], [], [], [], [], {})

        row, = rows
        assert row.chars == []
        assert row.cols == []
        assert row.positions == []
        assert row.posfrom == 0
        assert row.posto == 0

    def test_newline(self):
        rows = screen.col_splitter(
            5, 0, '0123\n', [1] * 5, list(range(5)), [1] * 5, [0] * 5, {})

        row, = rows
        assert row.chars == '0123\n'
        assert row.cols == [1, 1, 1, 1, 1]
        assert row.positions == [0, 1, 2, 3, 4]
        assert row.posfrom == 0
        assert row.posto == 5

    def test_controlchr(self):
        chrs, cols, positions, intervals = screen.translate_chars(
            0, '\x01\x02\x03\x7f', 8, 2)
        rows = screen.col_splitter(
            6, 0, chrs, cols, positions, intervals, [0] * 4, {})

        row1, row2, row3, row4 = rows
        assert row1.chars == '\\x01\\'
        assert row2.chars == 'x02\\x'
        assert row3.chars == '03\\x7'
        assert row4.chars == 'f'

    def test_nowrap(self):
        class C:
            nowrap = True

        rows = screen.col_splitter(4, 0, '0123456789', [1] * 10,
                                   list(range(10)), [0] * 10,
                                   [0, 0, 1, 1, 1, 0, 0, 0, 0, 0], {1: C})

        row1, row2, row3, row4 = rows
        assert row1.chars == '01'
        assert row2.chars == '234'
        assert row3.chars == '567'
        assert row4.chars == '89'

        rows = screen.col_splitter(3, 0, '0123456789', [1] * 10,
                                   list(range(10)), [0] * 10,
                                   [1, 1, 1, 1, 1, 0, 0, 0, 0, 0], {1: C})

        row1, row2, row3, row4, row5 = rows
        assert row1.chars == '01'
        assert row2.chars == '23'
        assert row3.chars == '45'
        assert row4.chars == '67'
        assert row5.chars == '89'

        rows = screen.col_splitter(3, 0, '0123456789', [1] * 10,
                                   list(range(10)), [0] * 10,
                                   [0, 0, 1, 1, 1, 0, 0, 0, 0, 0], {1: C})

        row1, row2, row3, row4, row5 = rows
        assert row1.chars == '01'
        assert row2.chars == '23'
        assert row3.chars == '45'
        assert row4.chars == '67'
        assert row5.chars == '89'

    def test_combine(self):
        rows = screen.col_splitter(
            3, 0, '0\u0300\u0300\u0300\u0300\u0300\u030012',
            [1, 0, 0, 0, 0, 0, 0, 1, 1], list(range(9)), [0] * 9, [0] * 9, {})

        row1, row2 = rows
        assert row1.chars == '0\u0300\u0300\u0300\u0300\u0300\u03001'
        assert row1.cols == [1, 0, 0, 0, 0, 0, 0, 1]
        assert row1.positions == [0, 1, 2, 3, 4, 5, 6, 7]
        assert row1.posfrom == 0
        assert row1.posto == 8

        assert row2.chars == '2'
        assert row2.cols == [1]
        assert row2.positions == [8]
        assert row2.posfrom == 8
        assert row2.posto == 9


class TestScreen(kaa_testutils._TestScreenBase):

    def test_build(self):
        scrn = self._getscreen("abc")

        row, = scrn.get_visible_rows()
        assert row.posfrom == 0
        assert row.posto == 3

    def test_nowrap(self):
        scrn = self._getscreen("abcdefg\n1234567890", width=3, nowrap=True)

        row1, row2 = scrn.get_visible_rows()
        assert row1.posfrom == 0
        assert row1.posto == 8

        assert row2.posfrom == 8
        assert row2.posto == 18

    def test_wrapindent(self):
        scrn = self._getscreen('    ' + '0123456789' * 16, width=80)
        row1, row2, row3 = scrn.get_visible_rows()
        assert row1.posfrom == 0
        assert row1.posto == 79

        assert row2.posfrom == 79
        assert row2.posto == 154

        assert row3.posfrom == 154
        assert row3.posto == 164

    def test_getrow(self):
        scrn = self._getscreen("abcdefghi", width=4, height=1)
        scrn.locate(6, top=True)

        idx, row = scrn.getrow(6)
        assert idx == 2
        assert row.posfrom == 6
        assert row.posto == 9

        idx, row = scrn.getrow(9)
        assert idx == 2
        assert row.posfrom == 6
        assert row.posto == 9

        scrn = self._getscreen("1\n2\n3\n", width=3, height=2)
        scrn.locate(4, top=True)

        idx, row = scrn.getrow(0)
        assert idx == -1
        assert row is None

        scrn = self._getscreen("1\n2\n3\n4\n4\n5\n6\n7\n", width=3, height=2)
        scrn.locate(4, top=True)

        idx, row = scrn.getrow(10)
        assert idx == -1
        assert row is None

        scrn = self._getscreen("abcdefghi\n", width=4, height=2)
        scrn.locate(6, top=True)

        idx, row = scrn.getrow(9)
        assert idx == 2
        assert row.posfrom == 6
        assert row.posto == 10

        idx, row = scrn.getrow(10)
        assert idx == 3
        assert row.posfrom == 10
        assert row.posto == 10

    def test_get_pos_under(self):
        scrn = self._getscreen("\u3042\u3043\u3044\n")
        assert scrn.get_pos_under(0, 3) == 1
        assert scrn.get_pos_under(1, 3) == 4
        assert scrn.get_pos_under(0, 100) == 3

        scrn = self._getscreen("0123456789", width=4)
        assert scrn.get_pos_under(1, 3) == 5
        assert scrn.get_pos_under(1, 10) == 5
        assert scrn.get_pos_under(3, 2) == 10

        scrn = self._getscreen("\t", width=4)
        assert scrn.get_pos_under(1, 0) == 1

        scrn = self._getscreen("\t\n", width=4)
        assert scrn.get_pos_under(1, 0) == 1
        assert scrn.get_pos_under(1, 10) == 1

    def test_get_pos_above(self):
        scrn = self._getscreen("\u3042\u3043\u3044\n")
        assert scrn.get_pos_above(0, 3) == 1
        assert scrn.get_pos_under(1, 3) == 4
        assert scrn.get_pos_under(0, 100) == 3

        scrn = self._getscreen("0123456789", width=4)
        assert scrn.get_pos_above(1, 3) == 5
        assert scrn.get_pos_above(1, 10) == 5
        assert scrn.get_pos_above(3, 2) == 10

        scrn = self._getscreen("\t\t", width=4)
        assert scrn.get_pos_above(1, 0) == 0

        scrn = self._getscreen("\t\t", width=3)
        assert scrn.get_pos_above(1, 0) == 0

    def test_getpos_fromrowcol(self):
        return
        scrn = self._getscreen("\u3042\u3043\u3044\n")
        assert scrn.getpos_fromrowcol(0, 3) == 1
        assert scrn.getpos_fromrowcol(1, 3) == 4
        assert scrn.getpos_fromrowcol(0, 100) == 3

        scrn = self._getscreen("0123456789", width=4)
        assert scrn.getpos_fromrowcol(1, 3) == 5
        assert scrn.getpos_fromrowcol(3, 2) == 10

        scrn = self._getscreen("\t\t\t\t", width=10)
        assert scrn.getpos_fromrowcol(0, 0) == 0
        assert scrn.getpos_fromrowcol(0, 1) == 0
        assert scrn.getpos_fromrowcol(0, 2) == 0
        assert scrn.getpos_fromrowcol(0, 3) == 0
        assert scrn.getpos_fromrowcol(0, 4) == 1
        assert scrn.getpos_fromrowcol(0, 5) == 1
        assert scrn.getpos_fromrowcol(0, 6) == 1
        assert scrn.getpos_fromrowcol(0, 7) == 1
        assert scrn.getpos_fromrowcol(0, 8) == 2
        assert scrn.getpos_fromrowcol(0, 9) == 2

        assert scrn.getpos_fromrowcol(1, 0) == 2
        assert scrn.getpos_fromrowcol(1, 1) == 2
        assert scrn.getpos_fromrowcol(1, 2) == 2
        assert scrn.getpos_fromrowcol(1, 3) == 3
        assert scrn.getpos_fromrowcol(1, 4) == 3

        scrn = self._getscreen('\t' * 2 + '01234567890', width=9)
        assert scrn.getpos_fromrowcol(0, 7) == 1

    def test_get_pos_at_cols(self):
        scrn = self._getscreen("012345\nabcdef", width=4, height=3)
        ret = scrn.get_pos_at_cols(7, 4)
        assert ret == 11

    def test_newline(self):
        scrn = self._getscreen("abc\n")

        row1, row2 = scrn.get_visible_rows()
        assert row1.posfrom == 0
        assert row1.posto == 4

        assert row2.posfrom == 4
        assert row2.posto == 4

        scrn = self._getscreen("abc\ndef", height=1)
        row1, = scrn.get_visible_rows()
        assert row1.posfrom == 0
        assert row1.posto == 4

        assert row2.posfrom == 4
        assert row2.posto == 4

    def test_empty(self):
        scrn = self._getscreen("")

        row, = scrn.get_visible_rows()
        assert row.posfrom == row.posto == 0

    def test_locate_empty(self):
        scrn = self._getscreen("", width=3, height=1)

        scrn.locate(0, top=True)
        assert scrn.portfrom == 0
        assert scrn.portto == 1

        scrn.locate(0, middle=True)
        assert scrn.portfrom == 0
        assert scrn.portto == 1

        scrn.locate(0, bottom=True)
        assert scrn.portfrom == 0
        assert scrn.portto == 1

    def test_locate(self):
        scrn = self._getscreen("012345\nabcdef", width=4, height=3)
        scrn.locate(7, top=True, align_always=True)
        assert scrn.portfrom == 0
        assert scrn.portto == 2
        assert scrn.pos == 7

        row1, row2 = scrn.get_visible_rows()
        assert row1.posfrom == 7
        assert row1.posto == 10

        assert row2.posfrom == 10
        assert row2.posto == 13

        scrn.locate(8, middle=True, align_always=True)
        assert scrn.portfrom == 1
        assert scrn.portto == 4
        assert scrn.pos == 3

        row1, row2, row3 = scrn.get_visible_rows()
        assert row1.posfrom == 3
        assert row1.posto == 7

        assert row2.posfrom == 7
        assert row2.posto == 10

        assert row3.posfrom == 10
        assert row3.posto == 13

        scrn.locate(8, bottom=True, align_always=True)
        assert scrn.portfrom == 0
        assert scrn.portto == 3
        assert scrn.pos == 0

        row1, row2, row3 = scrn.get_visible_rows()
        assert row1.posfrom == 0
        assert row1.posto == 3

        assert row2.posfrom == 3
        assert row2.posto == 7

        assert row3.posfrom == 7
        assert row3.posto == 10

    def test_locate_build_entire_rows(self):
        scrn = self._getscreen("012345\nabcdef\n01234", width=10, height=1)
        scrn.build_entire_rows = True
        scrn.locate(0, top=True, refresh=True)
        assert len(scrn.rows) == 3

    def test_linedown(self):
        scrn = self._getscreen("", height=1)
        assert scrn.linedown() == False

        scrn = self._getscreen("abc", height=1)
        assert scrn.linedown() == False

        scrn = self._getscreen("abc\n", height=1)
        assert scrn.linedown()

        row, = scrn.get_visible_rows()
        assert scrn.pos == row.posfrom == 4
        assert row.posto == 4

        scrn = self._getscreen("abc\ndef", height=1)
        assert scrn.linedown()

        row, = scrn.get_visible_rows()
        assert scrn.pos == row.posfrom == 4
        assert row.posto == 7

        scrn = self._getscreen("abcdef", width=4, height=1)
        assert scrn.linedown()

        row, = scrn.get_visible_rows()
        assert scrn.pos == row.posfrom == 3
        assert row.posto == 6

    def test_lineup(self):
        scrn = self._getscreen("", height=1)
        assert scrn.lineup() == False

        scrn = self._getscreen("abc\ndef\n", height=1)
        scrn.linedown()
        assert scrn.pos == 4

        assert scrn.lineup()
        assert scrn.pos == 0
        row, = scrn.get_visible_rows()
        assert row.posfrom == 0
        assert row.posto == 4

        scrn = self._getscreen("abcdef", width=3, height=1)
        scrn.linedown()
        assert scrn.pos == 2

        assert scrn.lineup()
        assert scrn.pos == 0
        row, = scrn.get_visible_rows()
        assert row.posfrom == 0
        assert row.posto == 2

    def test_updatescreen(self):
        scrn = self._getscreen("abcdefg\nabcdefg\n", width=4, height=3)
        scrn.locate(3, top=True, refresh=True)

        scrn.document.insert(0, '123')
        scrn.on_document_updated(0, 3, 0)
        scrn.apply_updates()

        assert scrn.pos == 6

        # delete chars of same line of tos row
        scrn.document.delete(0, 3)
        scrn.on_document_updated(0, 0, 3)
        scrn.apply_updates()

        assert scrn.pos == 6

        scrn.locate(8, top=True, refresh=True)

        # delete chars of previous line of tos
        scrn.document.delete(0, 3)
        scrn.on_document_updated(0, 0, 3)
        scrn.apply_updates()

        assert scrn.pos == 5



import sys
import os
import curses
import subprocess
import kaa
from kaa.command import Commands, command, is_enable, norec, norerun


class ToolCommands(Commands):

    @command('python.console')
    @norerun
    def pythonconsole(self, wnd):
        from kaa.ui.pyconsole import pythonconsolemode
        pythonconsolemode.show_console()

    @command('python.debugger.run')
    @norerun
    def pythonchilddebugger(self, wnd):
        from kaa.ui.pythondebug import port
        port.run()

    @command('python.debugger.server')
    @norerun
    def pythondebuggerserver(self, wnd):
        from kaa.ui.pythondebug import port
        port.init_server()

    @command('command.rerun')
    @norerun
    def reruncommand(self, wnd):
        mode = wnd.document.mode
        for commandid in kaa.app.lastcommands:
            is_available, command = mode.get_command(commandid)
            if not command:
                msg = 'command {!r} is not registered.'.format(commandid)
                kaa.app.messagebar.set_message(msg)
                kaa.log.error(msg)
                return
            command(wnd)

    @command('edit.paste-lines')
    @norerun
    def edit_pastelines(self, wnd):
        from kaa.ui.pastelines import pastelinesmode
        doc = pastelinesmode.PasteLinesMode.build(wnd)

        kaa.app.show_dialog(doc)

    @command('tools.execute-shell-command')
    @norerun
    def execute_shell_command(self, wnd):
        def callback(w, s):
            s = s.strip()
            if s:
                kaa.app.config.hist('shellcommand').add(s)
                ret = subprocess.check_output(
                    s, stderr=subprocess.STDOUT,
                    shell=True,
                    universal_newlines=True)

                wnd.document.mode.put_string(wnd, ret)
                wnd.screen.selection.clear()

                kaa.app.messagebar.set_message(
                    "{} letters inserted".format(len(ret)))

        hist = [s for s, info in kaa.app.config.hist('shellcommand').get()]
        from kaa.ui.inputline import inputlinemode
        doc = inputlinemode.InputlineMode.build('Shell command:',
                                                callback, history=hist)
        kaa.app.messagebar.set_message('Execute shell command')

        kaa.app.show_dialog(doc)

    def _exec_cmd(self, cmd):
        # todo: move to util
        import select
        import errno
        # http://stackoverflow.com/a/12207447
        # This page sujests bare os pipes is never buffered.
        # Is it true? Looks works fine, though.
        master, slave = os.pipe()
        # todo: Make these handles not inherited in Python 3.3.
        with subprocess.Popen(cmd, shell=True, stdout=slave,
                              stderr=slave, bufsize=1,
                              universal_newlines=True) as process:

            os.close(slave)

            output = os.fdopen(master)
            try:
                lines = []
                while process.poll() is None:
                    rlist, _, _ = select.select([master], [], [])
                    for f in rlist:
                        l = output.readline()
                        lines.append(l)
                        print(l, end='')
            finally:
                output.close()

            retcode = process.poll()
            if retcode:
                code = errno.errorcode.get(retcode)
                if code:
                    code = '(%s):' % code
                else:
                    code = ''

                msg = os.strerror(retcode)
                kaa.app.messagebar.set_message(
                    '{code}{msg}'.format(
                        code=code, msg=msg, retcode=retcode))

            return ''.join(lines)

    @command('tools.make')
    @norerun
    def execute_make(self, wnd):
        def callback(w, s):
            s = s.strip()
            if s:
                kaa.app.config.hist('makecommand').add(s)
                # todo: move these lines to kaa.cui.*
                curses.def_prog_mode()
                curses.endwin()
                try:
                    ret = self._exec_cmd(s)
                finally:
                    curses.reset_prog_mode()
                    wnd.mainframe.refresh()

                from kaa.ui.makeoutput import makeoutputmode
                makeoutputmode.show(s, ret)

        from kaa.ui.inputline import inputlinemode
        hist = [s for s, info in kaa.app.config.hist('makecommand').get()]
        value = 'make' if not hist else hist[0]
        doc = inputlinemode.InputlineMode.build('Make command:', callback,
                                                history=hist, value=value)
        kaa.app.messagebar.set_message('Execute command')

        kaa.app.show_dialog(doc)



import kaa
from . import wnd, editor


class _dialogwnd(wnd.Window, kaa.context.ContextRoot):
    focused_child = None

    def __init__(self, parent, pos=None):

        if pos is None:
            # specify default position to avoid flicker
            pos = self._get_defaultpos(parent)
        super().__init__(parent, pos=pos)

        self.set_label('popup', self)

    def _get_defaultpos(self, parent):
        if not parent:
            parent = self.mainframe
        l, t, r, b = parent.rect
        return (max(0, l), max(0, b - 2))

    def destroy(self):
        # Am I a active dialog?
        update_focus = kaa.app.focus in self.walk_children()
        super().destroy()

        # update focus
        if update_focus:
            kaa.app.set_focus(None)

        try:
            index = self.mainframe.popups.index(self)
        except ValueError:
            pass
        else:
            del self.mainframe.popups[index]
            if update_focus:
                if index > 0:
                    kaa.app.set_focus(self.mainframe.popups[index - 1])
                elif not self.mainframe.closed:
                    kaa.app.set_focus(self.mainframe)

        self.destroy_context()

        # If this dialog was inputline window, resize entire screen now.
        if self.mainframe.inputline is self:
            self.mainframe.inputline = None
            self.mainframe.on_console_resized()

    def draw_screen(self):
        pass

    def on_focus(self):
        """Got focus"""
        if self.parent and not self.parent.closed:
            self.parent.focused_child = self


class DialogWnd(_dialogwnd):

    def __init__(self, parent, doc, pos=None):
        super().__init__(parent, pos)
        self.inputs = []
        self.input = editor.TextEditorWindow(parent=self)
        self.set_label('editor', self.input)
        self.input.show_doc(doc)

    def destroy(self):
        self.input = None
        self.inputs = []
        super().destroy()

    def activate(self):
        for input in self.inputs:
            input.bring_top()
        super().activate()

    def on_focus(self):
        super().on_focus()
        if self.focused_child and not self.focused_child.closed:
            self.focused_child.activate()
        else:
            self.input.activate()

    def on_console_resized(self):
        """Resize wnds"""

        rc_input, rects = self.input.document.mode.resize_inputs(
            self.input, self.inputs)
        l, t, r, b = rc_input
        if rects:
            t = min(t, rects[0][1])
            b = max(b, rects[-1][3])
        self.set_rect(l, t, r, b)

        for w, (l, t, r, b) in zip(self.inputs, rects):
            w.set_rect(l, t, r, b)

        self.input.set_rect(*rc_input)

    def start(self):
        self.activate()
        self.input.document.mode.on_start(self.input)

    def add_doc(self, label, pos, doc):
        input = editor.TextEditorWindow(parent=self)
        self.set_label(label, input)
        input.show_doc(doc)
        self.inputs.insert(pos, input)
        return input



import kaa
from kaa.filetype.default import modebase
from kaa.theme import Theme, Style

MessageBarThemes = {
    'basic': [
        Style('default', 'Base3', 'default'),
        Style('rec', 'Base3', 'red'),
    ],
}


class MessageBarMode(modebase.ModeBase):
    USE_UNDO = False
    message = ''

    def init_themes(self):
        super().init_themes()
        self.themes.append(MessageBarThemes)

    def on_set_document(self, doc):
        super().on_set_document(doc)
        doc.undo = None

    def set_message(self, msg):
        if msg != self.message:
            self.message = msg
            self.update()

    def update(self):
        style_default = self.get_styleid('default')
        style_rec = self.get_styleid('rec')

        self.document.delete(0, self.document.endpos())
        if kaa.app.macro.is_recording():
            self.document.append('[REC]', style_rec)
            self.document.append(' ', style_default)

        self.document.append(self.message, style_default)



import os
import kaa
from kaa import document
from kaa.filetype.default import defaultmode
from kaa.ui.dialog import dialogmode
from kaa.theme import Theme, Style, Overlay
from kaa.command import command, norec, norerun
from kaa.keyboard import *
from kaa.commands import (appcommand, filecommand)
from kaa.ui.selectlist import selectlist
import kaa.ui.pythondebug.port
from kaa.ui.inputline import inputlinemode


def show_expr(port, depth):
    def callback(w, s):
        s = s.strip()
        if s:
            kaa.app.config.hist('pythondebug_expr').add(s)
            port.show_expr(depth, s)

    hist = [s for s, info in kaa.app.config.hist('pythondebug_expr').get()]
    doc = inputlinemode.InputlineMode.build('Python expression:',
                                            callback, history=hist)
    kaa.app.show_dialog(doc)


breakpoints_keys = {
    'd': 'breakpoints.delete',
    (ctrl, 'h'): 'breakpoints.delete',
    delete: 'breakpoints.delete',
    backspace: 'breakpoints.delete',
}


class BreakPoints(selectlist.SelectItemList):
    SEP = '\n'
    caption = 'Break points'

    @classmethod
    def build(cls, port):
        doc = super().build()
        doc.mode.show_breakpoints(port)
        kaa.app.messagebar.set_message("Hit 'd' key to remove breakpoint")
        return doc

    def init_keybind(self):
        super().init_keybind()
        self.keybind.add_keybind(breakpoints_keys)

    def show_breakpoints(self, port):
        self.port = port
        breakpoints = self.port.get_breakpoints()
        breakpoints.sort(key=lambda o: (o.filename, o.lineno))
        items = []
        for bp in breakpoints:
            caption = '{bp.filename}:{bp.lineno}'.format(bp=bp)
            c = selectlist.SelectItem(
                'selectitem', 'selectitem-active', caption, bp)
            items.append(c)

        self.update_doc(items)

    def on_esc_pressed(self, wnd, event):
        super().on_esc_pressed(wnd, event)
        popup = wnd.get_label('popup')
        popup.destroy()
        kaa.app.messagebar.set_message('')

    @command('breakpoints.delete')
    @norec
    @norerun
    def delete_bp(self, wnd):
        if not self.cursel:
            return
        bp = self.cursel.value
        for doc in document.Document.all:
            if doc.get_filename() == bp.filename:
                if bp in doc.marks:
                    del doc.marks[bp]
                break

        index = self.items.index(self.cursel)
        items = [item for item in self.items if item is not self.cursel]
        self.update_doc(items)
        if not self.items:
            sel = None
        elif index < len(self.items):
            sel = self.items[index]
        else:
            sel = self.items[-1]

        self.update_sel(wnd, sel)

DebugThemes = {
    'basic': [
        Style('line', None, None),
        Style('filename', 'Blue', None),
        Style('lineno', 'Blue', None),
        Style('funcname', 'Blue', None),
        Style('dirname', 'Blue', None),
        Style('status', 'Base3', 'Red', nowrap=True),
        Overlay('current_stack', 'Black', 'Yellow'),
    ],
}


callstack_keys = {
    down: 'callstack_keys.next',
    (ctrl, 'n'): 'callstack_keys.next',
    (ctrl, 'f'): 'callstack_keys.next',
    tab: 'callstack_keys.next',
    up: 'callstack_keys.prev',
    (ctrl, 'p'): 'callstack_keys.prev',
    (ctrl, 'b'): 'callstack_keys.prev',
    (shift, tab): 'callstack_keys.prev',
}


class PythonCallStackMode(dialogmode.DialogMode):
    SHOW_CURSOR = False
    autoshrink = True
    cursel = 0

    def setup(self):
        super().setup()
        self.updated_wnds = set()

    def init_themes(self):
        super().init_themes()
        self.themes.append(DebugThemes)

    def init_keybind(self):
        super().init_keybind()
        self.keybind.add_keybind(callstack_keys)

    def init_commands(self):
        super().init_commands()

        self.app_commands = appcommand.ApplicationCommands()
        self.register_command(self.app_commands)

        self.file_commands = filecommand.FileCommands()
        self.register_command(self.file_commands)

    def close(self):
        self._restore_highlight()
        super().close()
        self.port = None

    def on_str(self, wnd, s):
        pass

    def calc_position(self, wnd):
        height = wnd.screen.get_total_height()
        height = min(height,
                     (wnd.mainframe.height - wnd.mainframe.MESSAGEBAR_HEIGHT) // 2)
        top = wnd.mainframe.height - height - wnd.mainframe.MESSAGEBAR_HEIGHT
        return 0, top, wnd.mainframe.width, top + height

    def build(self, stack):
        self.document.marks.clear()
        self.document.delete(0, self.document.endpos())

        self.stack = tuple(stack or ())
        with dialogmode.FormBuilder(self.document) as f:
            f.append_text('button', '[&Step]',
                          shortcut_style='button.shortcut',
                          on_shortcut=self.on_step)

            f.append_text('button', '[&Next]',
                          shortcut_style='button.shortcut',
                          on_shortcut=self.on_next)

            f.append_text('button', '[&Return]',
                          shortcut_style='button.shortcut',
                          on_shortcut=self.on_return)

            f.append_text('button', '[&Continue]',
                          shortcut_style='button.shortcut',
                          on_shortcut=self.on_continue)

            f.append_text('button', '[&Expr]',
                          shortcut_style='button.shortcut',
                          on_shortcut=self.on_expr)

            f.append_text('button', '[&Breakpoint]',
                          shortcut_style='button.shortcut',
                          on_shortcut=self.on_breakpoint)

            f.append_text('button', '[&Quit]',
                          shortcut_style='button.shortcut',
                          on_shortcut=self.on_quit)

            f.append_text('status', '-Waiting-', mark_pair='status')

            f.append_text('default', '\n')

            for n, (fname, lno, funcname, lines) in enumerate(self.stack):
                s = self.document.endpos()

                dirname, filename = os.path.split(fname)
                f.append_text('filename', filename.replace('&', '&&'))
                f.append_text('default', ':')

                f.append_text('lineno', str(lno))
                f.append_text('default', ':')

                f.append_text('funcname', funcname.replace('&', '&&'))
                f.append_text('default', ':')

                f.append_text('dirname', dirname.replace('&', '&&') + '\n')

                line = (lines[0] if lines else '').strip()
                if not line:
                    line = '(empty line)'
                f.append_text('line', line.replace('&', '&&') + '\n')

                t = self.document.endpos()
                self.document.marks[('stack', n)] = (s, t)


    def update(self, wnd, stack):
        self.build(stack)
        self.update_sel(wnd, 0)
        wnd.mainframe.on_console_resized()

    def update_sel(self, wnd, n):
        if self.cursel != n:
            f, t = self.document.marks.get(
                ('stack', self.cursel), (None, None))
            if f is not None:
                wnd.set_line_overlay(f, None)

        f, t = self.document.marks.get(('stack', n), (None, None))
        if f is not None:
            wnd.set_line_overlay(f, 'current_stack')

        self.cursel = n
        if n < len(self.stack):
            (fname, lno, funcname, lines) = self.stack[n]
            self.show_line(fname, lno)
            wnd.activate()

            if f is not None:
                wnd.screen.locate(f, middle=True)

    def on_esc_pressed(self, wnd, event):
        super().on_esc_pressed(wnd, event)
        popup = wnd.get_label('popup')
        popup.destroy()
        kaa.app.messagebar.set_message('')

    def _restore_highlight(self):
        for w in self.updated_wnds:
            w.clear_line_overlay()
        self.updated_wnds = set()

    def _locate_doc(self, wnd, doc, lineno):
        pos = doc.get_lineno_pos(lineno)
        tol = doc.gettol(pos)

        wnd.clear_line_overlay()
        wnd.set_line_overlay(tol, 'current_row')

        wnd.screen.locate(tol, middle=True, align_always=True)
        wnd.cursor.setpos(tol, middle=True)
        wnd.activate()
        self.updated_wnds.add(wnd)

    def show_line(self, filename, lineno):
        filename = os.path.abspath(filename)
        doc = document.Document.find_filename(filename)
        if doc:
            wnd = doc.wnds[0]
            wnd.activate()
        else:
            doc = kaa.app.storage.openfile(filename, nohist=True)
            wnd = kaa.app.show_doc(doc)

        self._locate_doc(wnd, doc, lineno)

    def set_status(self, wnd, status):
        f, t = self.document.marks['status']
        styleid = self.get_styleid('status')
        self.document.replace(f, t, status, styleid)
        wnd.clear_line_overlay()
        self.update_sel(wnd, 0)

    @command('callstack_keys.prev')
    @norec
    @norerun
    def stack_prev(self, wnd):
        if self.cursel:
            self.update_sel(wnd, self.cursel - 1)

    @command('callstack_keys.next')
    @norec
    @norerun
    def stack_next(self, wnd):
        if self.cursel < len(self.stack) - 1:
            self.update_sel(wnd, self.cursel + 1)

    def on_step(self, wnd):
        self.port.set_step()

    def on_next(self, wnd):
        self.port.set_next()

    def on_return(self, wnd):
        self.port.set_return()

    def on_continue(self, wnd):
        self.port.set_continue()

    def on_expr(self, wnd):
        show_expr(self.port, self.cursel)

    def on_breakpoint(self, wnd):
        doc = BreakPoints.build(self.port)
        kaa.app.show_dialog(doc)

    def on_quit(self, wnd):
        self.port.close()


def show_callstack(port, stack):
    # update current breakpoints
    kaa.ui.pythondebug.port.update_breakpoints()

    buf = document.Buffer()
    doc = document.Document(buf)
    doc.set_title('Python call stack')
    mode = PythonCallStackMode()
    mode.port = port

    doc.setmode(mode)
    mode.build(stack)

    dlg = kaa.app.show_inputline(doc)
    ret = dlg.get_label('editor')
    mode.update_sel(ret, 0)
    return ret



import kaa_testutils
from kaa import highlight


class TestKeyword(kaa_testutils._TestDocBase):

    def test_keyword(self):
        kwds = highlight.Keywords(
            'keywords', 'keyword', ['if', 'while', 'for'])
        tokenizer = highlight.Tokenizer([kwds])

        doc = self._getdoc('if while for ')
        hl = highlight.Highlighter(tokenizers=[tokenizer])

        assert [
            (0, 2, kwds.tokenid),
            (2, 3, hl.tokenizers[0].nulltoken),
            (3, 8, kwds.tokenid),
            (8, 9, hl.tokenizers[0].nulltoken),
            (9, 12, kwds.tokenid),
            (12, 13, hl.tokenizers[0].nulltoken)
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

    def test_keyword_resume(self):
        kwds = highlight.Keywords('keywords', 'style', ['if', 'while', 'for'])
        tokenizer = highlight.Tokenizer([kwds])
        hl = highlight.Highlighter(tokenizers=[tokenizer])

        doc = self._getdoc('   while')
        doc.styles.setints(3, 8, kwds.tokenid)

        assert 3 == kwds.resume_pos(hl, tokenizer, doc, 5)

    def test_span(self):
        span = highlight.Span('str', 'style', '"""', '"""')
        tokenizer = highlight.Tokenizer([span])
        hl = highlight.Highlighter(tokenizers=[tokenizer])

        doc = self._getdoc(' """ """ ')
        ret = list((f, t, style) for f, t, style in hl.highlight(doc, 0))
        assert [
            (0, 1, hl.tokenizers[0].nulltoken),
            (1, 4, span.span_start),
            (4, 5, span.span_mid),
            (5, 8, span.span_end),
            (8, 9, hl.tokenizers[0].nulltoken), ] == ret

        span = highlight.Span('str', 'style', '"', '"')
        tokenizer = highlight.Tokenizer([span])

        hl = highlight.Highlighter(tokenizers=[tokenizer])

        doc = self._getdoc('"" ""')
        ret = list((f, t, style) for f, t, style in hl.highlight(doc, 0))
        assert [
            (0, 1, span.span_start),
            (1, 2, span.span_end),
            (2, 3, hl.tokenizers[0].nulltoken),
            (3, 4, span.span_start),
            (4, 5, span.span_end), ] == ret

    def test_span_escape(self):
        span = highlight.Span('str', 'style', '"""', '"""', '\\')
        tokenizer = highlight.Tokenizer([span])

        hl = highlight.Highlighter(tokenizers=[tokenizer])

        doc = self._getdoc(r'"""\""""')
        ret = list((f, t, style) for f, t, style in hl.highlight(doc, 0))
        assert [
            (0, 3, span.span_start),
            (3, 5, span.span_mid),
            (5, 8, span.span_end), ] == ret

    def test_span_noclose(self):
        kwds = highlight.Span('str', 'style', '"', '"', '\\')
        tokenizer = highlight.Tokenizer([kwds])
        hl = highlight.Highlighter(tokenizers=[tokenizer])

        doc = self._getdoc(r'"\"abcdefg')
        ret = list((f, t, style) for f, t, style in hl.highlight(doc, 0))
        assert [
            (0, 1, kwds.span_start),
            (1, 10, kwds.span_mid), ] == ret

    def test_span_resume(self):
        span = highlight.Span('str', 'style', '"', '"', None)
        tokenizer = highlight.Tokenizer([span])
        hl = highlight.Highlighter(tokenizers=[tokenizer])

        doc = self._getdoc(r'   "TEXT"')
        doc.styles.setints(3, 9, span.span_start)

        assert 3 == span.resume_pos(hl, tokenizer, doc, 5)

    def test_subsection(self):

        kwd = highlight.Keywords('keywords', 'style', ['abcdefg'])
        sub = highlight.Tokenizer([kwd])

        brace = highlight.SubSection('sub', 'style', r'\(', sub)
        tokenizer = highlight.Tokenizer([brace])

        hl = highlight.Highlighter(tokenizers=[tokenizer, sub])

        doc = self._getdoc(r'   (abcdefg   ')
        ret = list((f, t, style) for f, t, style in hl.highlight(doc, 0))
        assert [
            (0, 3, hl.tokenizers[0].nulltoken),
            (3, 4, brace.section_start),
            (4, 11, kwd.tokenid),
            (11, 14, hl.tokenizers[1].nulltoken)] == ret

    def test_subsection_resume(self):
        sub = highlight.Tokenizer([])

        brace = highlight.SubSection('sub', 'style', r'\(', sub)
        tokenizer = highlight.Tokenizer([brace])
        hl = highlight.Highlighter(tokenizers=[tokenizer])

        doc = self._getdoc(r'   (')
        doc.styles.setints(3, 4, brace.section_start)

        assert 3 == brace.resume_pos(hl, tokenizer, doc, 3)

    def test_endsection(self):
        end = highlight.EndSection('sub', 'style', r'\)')
        tokenizer = highlight.Tokenizer([end])
        hl = highlight.Highlighter(tokenizers=[tokenizer])

        doc = self._getdoc(r'   )')
        doc.styles.setints(3, 4, end.section_end)

        assert 3 == end.resume_pos(hl, tokenizer, doc, 3)

    def test_endsection_resume(self):
        end = highlight.EndSection('sub', 'style', r'\)')
        tokenizer = highlight.Tokenizer([end])
        hl = highlight.Highlighter(tokenizers=[tokenizer])

        doc = self._getdoc(r'   )')
        doc.styles.setints(3, 4, end.section_end)

        assert 3 == end.resume_pos(hl, tokenizer, doc, 3)

    def test_subtokenizer(self):
        kwds = highlight.Keywords('keywords', 'keyword', ['a', 'b'])
        tokenizer1 = highlight.Tokenizer([kwds])
        sub1 = highlight.SubTokenizer('sub', 'a', tokenizer1)
        tokenizer = highlight.Tokenizer([sub1])
        hl = highlight.Highlighter(tokenizers=[tokenizer])

        doc = self._getdoc(r'a b c a b c')
        ret = list((f, t, style) for f, t, style in hl.highlight(doc, 0))
        assert [
            (0, 1, kwds.tokenid),
            (1, 2, tokenizer1.nulltoken),
            (2, 3, kwds.tokenid),
            (3, 6, tokenizer1.nulltoken),
            (6, 7, kwds.tokenid),
            (7, 8, tokenizer1.nulltoken),
            (8, 9, kwds.tokenid),
            (9, 11, tokenizer1.nulltoken),
        ] == ret


class TestSection:

    def test_walk(self):
        section1 = highlight.Section(None, None)
        section2 = highlight.Section(None, None)
        section3 = highlight.Section(None, None)
        section4 = highlight.Section(None, None)
        section5 = highlight.Section(None, None)
        section6 = highlight.Section(None, None)
        section7 = highlight.Section(None, None)
        section8 = highlight.Section(None, None)
        section9 = highlight.Section(None, None)
        section10 = highlight.Section(None, None)

        section1.add(section2)
        section1.add(section3)

        section3.add(section4)
        section3.add(section5)

        section5.add(section6)
        section5.add(section7)

        section1.add(section8)
        section1.add(section9)
        section9.add(section10)

        assert [section1, section2, section3, section4,
                section5, section6, section7, section8,
                section9, section10, ] == list(section1.walk())

    def _newsection(self, start):
        section = highlight.Section(start, None)
        return section

    def test_getsection(self):
        section1 = self._newsection(0)
        section2 = self._newsection(1)
        section3 = self._newsection(2)

        section1.add(section2)
        section2.add(section3)

        assert section1.get_section(0) is section1
        assert section1.get_section(1) is section2
        assert section1.get_section(2) is section3

    def test_deleteafter(self):
        section1 = self._newsection(0)
        section2 = self._newsection(1)
        section3 = self._newsection(2)
        section4 = self._newsection(3)
        section5 = self._newsection(4)
        section6 = self._newsection(5)
        section7 = self._newsection(6)
        section8 = self._newsection(6)
        section9 = self._newsection(6)
        section10 = self._newsection(7)
        section11 = self._newsection(8)
        section12 = self._newsection(9)
        section13 = self._newsection(10)

        section1.add(section2)
        section2.add(section3)
        section2.add(section4)
        section2.add(section5)
        section1.add(section6)
        section1.add(section7)
        section7.add(section8)
        section8.add(section9)
        section8.add(section10)
        section8.add(section11)
        section8.add(section12)
        section7.add(section12)

        assert section1.get_section(2) is section3
        aa = section1.get_section(8)
        assert section1.get_section(8) is section11

        section1.delete_after(8)
        assert section7.children == [section8]
        assert section1.children == [section2, section6, section7]

        section1.delete_after(3)
        assert section2.children == [section3, section4]
        assert section1.children == [section2]


class TestHighlight(kaa_testutils._TestDocBase):

    def get_highliter(self):
        kwds = highlight.Keywords("jskeyword", 'style', ['if', 'while'])
        lit = highlight.Span('string', 'style', '"', '"', escape='\\')
        endsection = highlight.EndSection('endscript', 'style', '</script>')
        jstoken = highlight.Tokenizer([kwds, lit, endsection])

        elem = highlight.Span('comment', 'style', '<!--', '-->')
        scriptelem = highlight.SubSection(
            'beginscript', 'style', '<script>', tokenizer=jstoken)
        htmltoken = highlight.Tokenizer([elem, scriptelem])

        hl = highlight.Highlighter(tokenizers=[htmltoken, jstoken])
        return hl, kwds, lit, endsection, elem, scriptelem

    def get_doc(self):
        return self._getdoc(r'abc<script>if while</script>if while<!-- comment -->')

    def test_highlight(self):

        hl, kwds, lit, endsection, elem, scriptelem = self.get_highliter()
        doc = self.get_doc()

        tokens = list(hl.highlight(doc, 0))
        assert [
            (0, 3, hl.tokenizers[0].nulltoken),
            (3, 11, scriptelem.section_start),
            (11, 13, kwds.tokenid),
            (13, 14, hl.tokenizers[1].nulltoken),
            (14, 19, kwds.tokenid),
            (19, 28, endsection.section_end),
            (28, 36, hl.tokenizers[0].nulltoken),
            (36, 40, elem.span_start),
            (40, 49, elem.span_mid),
            (49, 52, elem.span_end),
        ] == tokens

    def test_resume(self):
        hl, kwds, lit, endsection, elem, scriptelem = self.get_highliter()
        doc = self.get_doc()
        list(hl.highlight(doc, 0))
        doc.styles.setints(14, 19, kwds.tokenid)

        assert 0 == hl.get_resume_pos(doc, 0)
        assert 0 == hl.get_resume_pos(doc, 1)
        assert 14 == hl.get_resume_pos(doc, 16)
        assert 19 == hl.get_resume_pos(doc, 28)
        assert 19 == hl.get_resume_pos(doc, 29)

        tokens = list(hl.highlight(doc, 19))
        assert [
            (19, 28, endsection.section_end),
            (28, 36, hl.tokenizers[0].nulltoken),
            (36, 40, elem.span_start),
            (40, 49, elem.span_mid),
            (49, 52, elem.span_end),
        ] == tokens

    def test_update(self):

        hl, kwds, lit, endsection, elem, scriptelem = self.get_highliter()
        doc = self.get_doc()
        doc.mode.highlight = hl
        doc.mode.highlight.update_style(doc)

        assert ([hl.tokenizers[0].nulltoken] * 3 + [scriptelem.section_start] * 8 + [kwds.tokenid] * 2 +
                [hl.tokenizers[1].nulltoken] + [kwds.tokenid] * 5 + [endsection.section_end] * 9 +
                [hl.tokenizers[0].nulltoken] * 8 + [elem.span_start] * 4 + [elem.span_mid] * 9 +
                [elem.span_end] * 3) == doc.styles.getints(0, len(doc.styles))

        doc.replace(16, 17, 'x')  # while -> whxle
        doc.mode.highlight.update_style(doc)

        assert ([hl.tokenizers[0].nulltoken] * 3 + [scriptelem.section_start] * 8 + [kwds.tokenid] * 2 +
                [hl.tokenizers[1].nulltoken] + [hl.tokenizers[1].nulltoken] * 5 + [endsection.section_end] * 9 +
                [hl.tokenizers[0].nulltoken] * 8 + [elem.span_start] * 4 + [elem.span_mid] * 9 +
                [elem.span_end] * 3) == doc.styles.getints(0, len(doc.styles))



import kaa
from kaa import document
from kaa.ui.dialog import dialogmode
from kaa.ui.selectlist import selectlist, filterlist
from kaa.command import command, norec, norerun
from kaa.keyboard import *

clipboardhist_keys = {
    '\r': 'clipboardhist.select',
    '\n': 'clipboardhist.select',
}


class ClipboardHist(filterlist.FilterListMode):
    MAX_CAPTION_LEN = 256
    SCREEN_NOWRAP = True

    def init_keybind(self):
        super().init_keybind()
        self.keybind.add_keybind(clipboardhist_keys)

    def set_candidates(self, candidates):
        self.candidates = []
        for c in candidates:
            caption = ' '.join(c.split())
            if not caption:
                caption = '(...)'
            if len(caption) > self.MAX_CAPTION_LEN:
                caption = caption[:self.MAX_CAPTION_LEN] + '...'
            c = selectlist.SelectItem(
                'selectitem', 'selectitem-active', caption, c)
            self.candidates.append(c)

    def _filter_items(self, query):
        if query:
            query = query.upper().split()
            items = []
            for s in self.candidates:
                u = s.value.upper()
                for q in query:
                    if q not in u:
                        break
                else:
                    items.append(s)
        else:
            items = self.candidates[:]

        sel = items[0] if items else None
        return sel, items

    def on_esc_pressed(self, wnd, event):

        super().on_esc_pressed(wnd, event)
        popup = wnd.get_label('popup')
        popup.destroy()
        kaa.app.messagebar.set_message("")

    @command('clipboardhist.select')
    @norec
    @norerun
    def on_selected(self, wnd):
        if self.cursel:
            self.target.document.mode.put_string(
                self.target, self.cursel.value)
            self.target.screen.selection.clear()
            kaa.app.clipboard.set(self.cursel.value)
            
        popup = wnd.get_label('popup')
        popup.destroy()
        kaa.app.messagebar.set_message("")


def show_history(wnd):
    words = list(kaa.app.clipboard.get_all())
    if not words:
        return

    doc = ClipboardHist.build()
    dlg = kaa.app.show_dialog(doc)
    doc.mode.target = wnd
    doc.mode.set_candidates(words)
    doc.mode.set_query(dlg.get_label('editor'), '')

    dlg.on_console_resized()
    kaa.app.messagebar.set_message("Hit tab/shift+tab to select text. ")

    return dlg



import kaa


class Theme:

    def __init__(self, styles):
        self.styles = {}
        self.overlays = {}
        self.add_styles(styles)

    def get_style(self, name):
        return self.styles[name]

    def add_styles(self, styles):
        for style in styles:
            if not isinstance(style, Overlay):
                # Style instance
                self.styles[style.name] = style
            else:
                self.overlays[style.name] = style

    def update(self, rhs):
        default = self.styles.get('default', None)
        self.add_styles(s.fill(default) for s in rhs)


class Style:

    def __init__(self, name, fgcolor, bgcolor, underline=False,
                 bold=False, nowrap=False, rjust=False):
        self.name = name
        self.fgcolor = fgcolor
        self.bgcolor = bgcolor
        self.underline = underline
        self.bold = bold
        self.nowrap = nowrap
        self.rjust = rjust

    def fill(self, default):
        return self.__class__(
            self.name,
            self.fgcolor if self.fgcolor is not None else default.fgcolor,
            self.bgcolor if self.bgcolor is not None else default.bgcolor,
            self.underline if self.underline is not None else default.underline,
            self.bold if self.bold is not None else default.bold,
            self.nowrap if self.nowrap is not None else default.nowrap,
            self.rjust if self.rjust is not None else default.rjust,
        )


class Overlay(Style):

    def __init__(self, name, fgcolor=None, bgcolor=None,
                 underline=None, bold=None):
        super().__init__(name, fgcolor, bgcolor, underline, bold)

    def fill(self, default):
        return self.__class__(
            self.name,
            self.fgcolor,
            self.bgcolor,
            self.underline,
            self.bold
        )



import argparse


def build_parser():
    parser = argparse.ArgumentParser(description='kaa text editor.')

    parser.add_argument('--version', dest='show_version', action='store_true',
                        default=False, help='show version info and exit')

    parser.add_argument('--no-init', dest='no_init', action='store_true',
                        default=False, help='skip loading initialization script')

    parser.add_argument('--init-script', dest='init_script',
                        help='execute file as initialization script instead of default initialization file')

    parser.add_argument('--palette', dest='palette', default=None,
                        help='color palette. available values: dark, light.')

    parser.add_argument('--term', '-t', dest='term', default='',
                        help='specify terminal type')

    parser.add_argument('file', nargs='*', default=[])

    return parser



import itertools
import re
import unicodedata
import kaa
from kaa.command import Commands, command, is_enable, norec, norerun
from kaa import document
from kaa.filetype.default import modebase
from gappedbuf import re as gre


class CursorCommands(Commands):

    @command('cursor.right')
    @norerun
    def right(self, wnd):
        if wnd.screen.selection.has_mark():
            wnd.cursor.right()
        elif wnd.screen.selection.is_selected():
            range = wnd.screen.selection.get_selrange()
            if wnd.cursor.pos > range[1]:
                wnd.cursor.right()
            else:
                wnd.cursor.setpos(wnd.cursor.adjust_nextpos(
                    wnd.cursor.pos, range[1]))
        else:
            wnd.document.mode.cancel_auto_indent(wnd)
            wnd.cursor.right()

        wnd.screen.selection.end_cursor()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @command('cursor.right.select')
    @norerun
    def right_select(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.begin_cursor(wnd.cursor.pos)
        wnd.cursor.right()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @command('cursor.left')
    @norerun
    def left(self, wnd):
        if wnd.screen.selection.has_mark():
            wnd.cursor.left()
        elif wnd.screen.selection.is_selected():
            range = wnd.screen.selection.get_selrange()
            if wnd.cursor.pos < range[0]:
                wnd.cursor.left()
            else:
                wnd.cursor.setpos(wnd.cursor.adjust_nextpos(
                    wnd.cursor.pos, range[0]))
        else:
            wnd.document.mode.cancel_auto_indent(wnd)
            wnd.cursor.left()

        wnd.screen.selection.end_cursor()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @command('cursor.left.select')
    @norerun
    def left_select(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.begin_cursor(wnd.cursor.pos)
        wnd.cursor.left()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @command('cursor.up')
    @norerun
    def up(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.end_cursor()
        wnd.cursor.up()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @command('cursor.up.select')
    @norerun
    def up_select(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.begin_cursor(wnd.cursor.pos)
        wnd.cursor.up()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @command('cursor.down')
    @norerun
    def down(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.end_cursor()
        wnd.cursor.down()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @command('cursor.down.select')
    @norerun
    def down_select(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.begin_cursor(wnd.cursor.pos)
        wnd.cursor.down()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @command('cursor.prev-line')
    @norerun
    def prev_line(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.end_cursor()
        wnd.cursor.prev_line()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @command('cursor.next-line')
    @norerun
    def next_line(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.end_cursor()
        wnd.cursor.next_line()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @command('cursor.word-right')
    @norerun
    def word_right(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.end_cursor()
        wnd.cursor.right(word=True)
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @command('cursor.word-right.select')
    @norerun
    def word_right_select(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.begin_cursor(wnd.cursor.pos)
        wnd.cursor.right(word=True)
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @command('cursor.word-left')
    @norerun
    def word_left(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.end_cursor()
        wnd.cursor.left(word=True)
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @command('cursor.word-left.select')
    @norerun
    def word_left_select(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.begin_cursor(wnd.cursor.pos)
        wnd.cursor.left(word=True)
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @command('cursor.pagedown')
    @norerun
    def pagedown(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.end_cursor()
        wnd.cursor.pagedown()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @command('cursor.pagedown.select')
    @norerun
    def pagedown_select(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.begin_cursor(wnd.cursor.pos)
        wnd.cursor.pagedown()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @command('cursor.pageup')
    @norerun
    def pageup(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.end_cursor()
        wnd.cursor.pageup()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @command('cursor.pageup.select')
    @norerun
    def pageup_select(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.begin_cursor(wnd.cursor.pos)
        wnd.cursor.pageup()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @command('cursor.home')
    @norerun
    def home(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.end_cursor()
        wnd.cursor.home()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @command('cursor.home.select')
    @norerun
    def home_select(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.begin_cursor(wnd.cursor.pos)
        wnd.cursor.home()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @command('cursor.end')
    @norerun
    def end(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.end_cursor()
        wnd.cursor.end()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @command('cursor.end.select')
    @norerun
    def end_select(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.begin_cursor(wnd.cursor.pos)
        wnd.cursor.end()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @command('cursor.top-of-line')
    @norerun
    def tol(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.end_cursor()
        wnd.cursor.tol(wnd.cursor.pos)
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @command('cursor.end-of-line')
    @norerun
    def eol(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.end_cursor()
        wnd.cursor.eol(wnd.cursor.pos)
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @command('cursor.top-of-file')
    @norerun
    def top(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.end_cursor()
        wnd.cursor.tof()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @command('cursor.top-of-file.select')
    @norerun
    def top_select(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.begin_cursor(wnd.cursor.pos)
        wnd.cursor.tof()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @command('cursor.end-of-file')
    @norerun
    def last(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.end_cursor()
        wnd.cursor.eof()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @command('cursor.end-of-file.select')
    @norerun
    def last_select(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.begin_cursor(wnd.cursor.pos)
        wnd.cursor.eof()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @command('cursor.go-to-line')
    @norerun
    @norec
    def go_to_line(self, wnd):
        def callback(w, s):
            s = s.strip()
            try:
                lineno = int(s)
            except ValueError as e:
                kaa.app.messagebar.set_message(str(e))
                return

            if lineno == 0 or lineno > wnd.document.buf.lineno.linecount():
                kaa.app.messagebar.set_message('Enter valid line number.')
                return

            wnd.document.mode.cancel_auto_indent(wnd)
            wnd.screen.selection.end_cursor()

            pos = wnd.document.get_lineno_pos(lineno)
            tol = wnd.document.gettol(pos)
            wnd.cursor.setpos(wnd.cursor.adjust_nextpos(wnd.cursor.pos, tol))

        from kaa.ui.inputline import inputlinemode
        doc = inputlinemode.InputlineMode.build('Line number:', callback,
                                                filter=inputlinemode.number_filter)
        kaa.app.messagebar.set_message("Enter line number")

        kaa.app.show_dialog(doc)


class ScreenCommands(Commands):

    @command('selection.begin-cursor')
    @norerun
    def selection_begin(self, wnd):
        wnd.screen.selection.begin_cursor(wnd.cursor.pos)

    @command('selection.set-to')
    @norerun
    def selection_set_end(self, wnd):
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @command('selection.end-cursor')
    @norerun
    def selection_clear(self, wnd):
        wnd.screen.selection.end_cursor()

    @command('selection.set-mark')
    @norerun
    def selection_set_mark(self, wnd):
        if not wnd.screen.selection.has_mark():
            wnd.screen.selection.set_mark(wnd.cursor.pos)
        else:
            wnd.screen.selection.set_mark(None)

    @command('selection.set-rectangle-mark')
    @norerun
    def selection_set_rectangle_mark(self, wnd):
        if not wnd.screen.selection.has_mark():
            wnd.screen.selection.set_rectangle_mark(wnd.cursor.pos)
        else:
            wnd.screen.selection.set_rectangle_mark(None)

    @command('selection.all')
    @norerun
    def select_all(self, wnd):
        f = wnd.cursor.adjust_nextpos(wnd.cursor.pos, 0)
        t = wnd.cursor.adjust_nextpos(wnd.cursor.pos, wnd.document.endpos())
        wnd.screen.selection.set_range(f, t)

    @command('selection.curline')
    @norerun
    def select_cur_line(self, wnd):
        tol = wnd.cursor.adjust_nextpos(
            wnd.cursor.pos,
            wnd.document.gettol(wnd.cursor.pos))
        eol = wnd.cursor.adjust_nextpos(
            wnd.cursor.pos,
            wnd.document.geteol(tol))

        wnd.screen.selection.set_range(tol, eol)

    @command('selection.curword')
    @norerun
    def select_cur_word(self, wnd):
        ret = wnd.document.mode.get_word_at(wnd.cursor.pos)
        if ret:
            f, t, cg = ret
            f = wnd.cursor.adjust_nextpos(wnd.cursor.pos, f)
            t = wnd.cursor.adjust_nextpos(wnd.cursor.pos, t)

            wnd.screen.selection.set_range(f, t)

    @command('selection.expand-sel')
    @norerun
    @norec
    def expand_sel(self, wnd):
        mode = wnd.document.mode
        keys = wnd.editmode.last_command_keys
        L = len(keys)

        if L >= 3 and keys[-1] == keys[-2] == keys[-3]:
            self.select_all(wnd)
            if kaa.app.macro.recording:
                kaa.app.macro.record(self.select_all)

        elif L >= 2 and (keys[-1] == keys[-2]):
            self.select_cur_line(wnd)
            if kaa.app.macro.recording:
                kaa.app.macro.record(self.select_cur_line)

        else:
            self.select_cur_word(wnd)
            if kaa.app.macro.recording:
                kaa.app.macro.record(self.select_cur_word)

    @command('screen.lineselection.begin')
    @norerun
    def lineselection_begin(self, wnd):
        tol = wnd.cursor.adjust_nextpos(wnd.document.gettol(wnd.cursor.pos))
        wnd.screen.selection.begin_cursor(tol)

    @command('screen.lineselection.set-end')
    @norerun
    def lineselection_set_end(self, wnd):
        f = wnd.screen.get_start()
        if f is None:
            self.select_cur_line(wnd)
        else:
            pos = wnd.cursor.pos
            if pos < f:
                tol = wnd.cursor.adjust_nextpos(
                    wnd.document.gettol(pos))
                wnd.screen.selection.set_end(tol)
            else:
                eol = wnd.cursor.adjust_nextpos(
                    wnd.document.geteol(pos))
                wnd.screen.selection.set_end(eol)


class EditCommands(Commands):
    def delete_sel(self, wnd):
        if wnd.screen.selection.is_selected():
            if not wnd.screen.selection.is_rectangular():
                sel = wnd.screen.selection.get_selrange()
                wnd.screen.selection.clear()
                if sel:
                    f, t = sel
                    wnd.document.mode.delete_string(wnd, f, t)
                    return True
            else:
                return self._delete_rect_sel(wnd)

    def _delete_rect_sel(self, wnd):
        if wnd.document.undo:
            wnd.document.undo.beginblock()
        try:
            (posfrom, posto, colfrom, colto
             ) = wnd.screen.selection.get_rect_range()

            while posfrom < posto:
                sel = wnd.screen.selection.get_col_string(
                    posfrom, colfrom, colto)
                if sel:
                    f, t, org = sel
                    if org.endswith('\n'):
                        t = max(f, t - 1)
                    wnd.document.mode.delete_string(wnd, f, t)
                    posto -= (t - f)
                posfrom = wnd.document.geteol(posfrom)

            return True
        finally:
            if wnd.document.undo:
                wnd.document.undo.endblock()

    @command('edit.delete')
    def delete(self, wnd):
        if self.delete_sel(wnd):
            return

        pos = wnd.cursor.pos
        nextpos = wnd.document.get_nextpos(pos)
        nextpos = wnd.cursor.adjust_nextpos(pos, nextpos)
        if pos < nextpos:
            wnd.document.mode.delete_string(wnd, pos, nextpos)

    @command('edit.delete.word')
    def delete_word(self, wnd):
        if self.delete_sel(wnd):
            return

        pos = wnd.cursor.pos
        wnd.cursor.right(word=True)
        nextpos = wnd.cursor.pos
        if pos < nextpos:
            wnd.document.mode.delete_string(wnd, pos, nextpos)

    @command('edit.delete.line')
    def delete_line(self, wnd):
        pos = wnd.cursor.pos
        nextpos = wnd.cursor.adjust_nextpos(
            pos, wnd.document.find_newline(pos))
        if pos < nextpos:
            wnd.document.mode.delete_string(wnd, pos, nextpos)

    @command('edit.delete.currentline')
    def delete_currentline(self, wnd):
        pos = wnd.cursor.pos
        f = wnd.cursor.adjust_nextpos(pos, wnd.document.gettol(pos))
        t = wnd.cursor.adjust_nextpos(pos, wnd.document.geteol(f))

        if f < t:
            wnd.document.mode.delete_string(wnd, f, t)

    @command('edit.backspace')
    def backspace(self, wnd):
        if self.delete_sel(wnd):
            return

        pos = wnd.cursor.pos
        prevpos = wnd.cursor.adjust_nextpos(
            pos, wnd.document.get_prevpos(pos))
        if prevpos < pos:
            if pos == wnd.screen.pos:
                # locate cursor before delete to scroll half page up
                wnd.cursor.setpos(prevpos)

            wnd.document.mode.delete_string(wnd, prevpos, pos)

    @command('edit.backspace.word')
    def backspace_word(self, wnd):
        if self.delete_sel(wnd):
            return

        pos = wnd.cursor.pos
        wnd.cursor.left(word=True)
        prevpos = wnd.cursor.pos
        if prevpos < pos:
            if prevpos < wnd.screen.pos:
                # locate cursor before delete to scroll half page up
                wnd.cursor.setpos(prevpos)

            wnd.document.mode.delete_string(wnd, prevpos, pos)

    @command('edit.newline')
    @norerun
    def newline(self, wnd):
        if not wnd.document.mode.auto_indent:
            wnd.document.mode.put_string(wnd, '\n')
            wnd.screen.selection.clear()
            return

        wnd.screen.selection.clear()
        wnd.document.mode.on_auto_indent(wnd)

    def _indent_line(self, wnd, pos):
        mode = wnd.document.mode
        f, t = mode.get_indent_range(pos)
        if pos > t:
            wnd.document.mode.put_string(wnd, '\t')
            wnd.screen.selection.clear()
            return

        if f != t:
            cols = mode.calc_cols(f, t)
        else:
            cols = 0

        s = mode.build_indent_str(cols + mode.indent_width)
        mode.replace_string(wnd, f, t, s, True)

    @command('edit.indent')
    def indent(self, wnd):
        if not wnd.screen.selection.is_selected():
            self._indent_line(wnd, wnd.cursor.pos)
            return

        doc = wnd.document
        tol, eol = doc.mode.get_line_sel(wnd)
        wnd.screen.selection.set_range(tol, eol)

        if wnd.document.undo:
            wnd.document.undo.beginblock()
        try:
            mode = wnd.document.mode
            while tol < wnd.screen.selection.get_end():
                f, t = mode.get_indent_range(tol)
                if f != t:
                    cols = mode.calc_cols(f, t)
                else:
                    cols = 0

                endpos = doc.endpos()
                eol = doc.geteol(tol)

                if (t + 1 < eol != endpos) or (t < eol == endpos):
                    s = mode.build_indent_str(cols + mode.indent_width)
                    mode.replace_string(wnd, f, t, s, False)
                    tol = eol + (len(s) - (t - f))
                else:
                    tol = eol
        finally:
            if wnd.document.undo:
                wnd.document.undo.endblock()

        f, t = wnd.screen.selection.get_selrange()
        wnd.cursor.setpos(f)
        wnd.cursor.savecol()

    def _dedent_line(self, wnd, pos):
        mode = wnd.document.mode
        f, t = mode.get_indent_range(pos)
        if f != t:
            cols = mode.calc_cols(f, t)
        else:
            cols = 0

        s = mode.build_indent_str(max(0, cols - mode.indent_width))
        mode.replace_string(wnd, f, t, s, True)

    @command('edit.dedent')
    def dedent(self, wnd):
        if not wnd.screen.selection.is_selected():
            self._dedent_line(wnd, wnd.cursor.pos)
            return

        doc = wnd.document
        tol, eol = doc.mode.get_line_sel(wnd)
        wnd.screen.selection.set_range(tol, eol)

        if wnd.document.undo:
            wnd.document.undo.beginblock()
        try:
            mode = wnd.document.mode
            while tol < wnd.screen.selection.get_end():
                f, t = mode.get_indent_range(tol)
                if f != t:
                    cols = mode.calc_cols(f, t)
                else:
                    cols = 0

                if cols:
                    s = mode.build_indent_str(max(0, cols - mode.indent_width))
                    mode.replace_string(wnd, f, t, s, False)

                tol = doc.geteol(tol)
        finally:
            if wnd.document.undo:
                wnd.document.undo.endblock()

        f, t = wnd.screen.selection.get_selrange()
        wnd.cursor.setpos(f)
        wnd.cursor.savecol()

    def _undo(self, wnd, rec):
        (action, args, kwargs) = rec
        if action == wnd.document.mode.UNDO_INSERT:
            pos, s, cur_pos, newpos = args
            wnd.document.delete(pos, pos + len(s))
            return cur_pos
        elif action == wnd.document.mode.UNDO_REPLACE:
            pos, posto, s, deled, cur_pos, newpos = args
            wnd.document.replace(pos, pos + len(s), deled)
            return cur_pos
        else:
            pos, posto, deled, cur_pos, newpos = args
            wnd.document.insert(pos, deled)

        return cur_pos

    @command('edit.undo')
    @norerun
    def undo(self, wnd):
        if wnd.document.undo and wnd.document.undo.can_undo():
            wnd.screen.selection.clear()
            pos = None
            for rec in wnd.document.undo.undo():
                pos = self._undo(wnd, rec)

            if pos is not None:
                wnd.cursor.setpos(pos)
                wnd.cursor.savecol()

            wnd.document.mode.on_edited(wnd)

    def _redo(self, wnd, rec):
        (action, args, kwargs) = rec
        if action == wnd.document.mode.UNDO_INSERT:
            pos, s, cur_pos, newpos = args
            wnd.document.insert(pos, s)
            return newpos
        elif action == wnd.document.mode.UNDO_REPLACE:
            pos, posto, s, deled, cur_pos, newpos = args
            wnd.document.replace(pos, posto, s)
            return pos
        else:
            pos, posto, deled, cur_pos, newpos = args
            wnd.document.delete(pos, posto)

        return newpos

    @command('edit.redo')
    @norerun
    def redo(self, wnd):
        if wnd.document.undo and wnd.document.undo.can_redo():
            wnd.screen.selection.clear()
            pos = None
            for rec in wnd.document.undo.redo():
                pos = self._redo(wnd, rec)

            if pos is not None:
                wnd.cursor.setpos(pos)
                wnd.cursor.savecol()

            wnd.document.mode.on_edited(wnd)

    def _get_sel(self, wnd):
        if wnd.screen.selection.is_selected():
            if not wnd.screen.selection.is_rectangular():
                f, t = wnd.screen.selection.get_selrange()
                return wnd.document.gettext(f, t)
            else:
                s = []
                (posfrom, posto, colfrom, colto
                 ) = wnd.screen.selection.get_rect_range()

                while posfrom < posto:
                    sel = wnd.screen.selection.get_col_string(
                        posfrom, colfrom, colto)
                    if sel:
                        f, t, org = sel
                        s.append(org.rstrip('\n'))
                    else:
                        s.append('')
                    posfrom = wnd.document.geteol(posfrom)

                return '\n'.join(s)

    def _copy_sel(self, wnd):
        sel = self._get_sel(wnd)
        if sel:
            kaa.app.clipboard.set(sel)
        return sel

    @command('edit.copy')
    def copy(self, wnd):
        self._copy_sel(wnd)
        wnd.screen.selection.set_mark(None)

    @command('edit.cut')
    def cut(self, wnd):
        if self._copy_sel(wnd):
            self.delete_sel(wnd)
        wnd.screen.selection.clear()

    @command('edit.paste')
    def paste(self, wnd):
        s = kaa.app.clipboard.get()
        if s:
            wnd.document.mode.put_string(wnd, s)
            wnd.screen.selection.clear()

    @command('edit.conv.upper')
    def conv_upper(self, wnd):
        s = self._get_sel(wnd)
        if s:
            wnd.document.mode.put_string(wnd, s.upper())
            wnd.screen.selection.clear()

    @command('edit.conv.lower')
    def conv_lower(self, wnd):
        s = self._get_sel(wnd)
        if s:
            wnd.document.mode.put_string(wnd, s.lower())
            wnd.screen.selection.clear()

    @command('edit.conv.nfkc')
    def conv_nfkc(self, wnd):
        s = self._get_sel(wnd)
        if s:
            wnd.document.mode.put_string(wnd, unicodedata.normalize('NFKC', s))
            wnd.screen.selection.clear()

    @command('edit.conv.full-width')
    def conv_fullwidth(self, wnd):
        import pyjf3
        s = self._get_sel(wnd)
        if s:
            wnd.document.mode.put_string(wnd, pyjf3.tofull(s))
            wnd.screen.selection.clear()

    @command('edit.word-complete')
    @norec
    @norerun
    def complete(self, wnd):
        from kaa.ui.wordcomplete import wordcompletemode
        wordcompletemode.show_wordlist(wnd)

    @command('edit.clipboard-history')
    @norec
    @norerun
    def clipboard_history(self, wnd):
        from kaa.ui.clipboardhist import clipboardhistmode
        clipboardhistmode.show_history(wnd)


class CodeCommands(Commands):

    @command('code.region.linecomment')
    def linecomment(self, wnd):
        if not wnd.screen.selection.is_selected():
            tol = wnd.document.gettol(wnd.cursor.pos)
            wnd.document.mode.insert_string(
                wnd, tol, wnd.document.mode.LINE_COMMENT,
                update_cursor=False)
            wnd.cursor.setpos(tol)
            wnd.cursor.savecol()
            return
        else:
            tol, eol = wnd.document.mode.get_line_sel(wnd)
            wnd.screen.selection.set_range(tol, eol)

            if wnd.document.undo:
                wnd.document.undo.beginblock()
            try:
                mode = wnd.document.mode
                while tol < wnd.screen.selection.get_end():
                    wnd.document.mode.insert_string(
                        wnd, tol, wnd.document.mode.LINE_COMMENT,
                        update_cursor=False)
                    tol = wnd.document.geteol(tol)
            finally:
                if wnd.document.undo:
                    wnd.document.undo.endblock()

            f, t = wnd.screen.selection.get_selrange()
            wnd.cursor.setpos(f)
            wnd.cursor.savecol()

    def _is_comment_line(self, wnd, pos):
        reobj = gre.compile(r'[ \t]*({})'.format(
            gre.escape(wnd.document.mode.LINE_COMMENT)))
        return reobj.match(wnd.document.buf, pos)

    @command('code.region.unlinecomment')
    def uncomment(self, wnd):
        if not wnd.screen.selection.is_selected():
            tol = wnd.document.gettol(wnd.cursor.pos)
            m = self._is_comment_line(wnd, tol)
            if m:
                f, t = m.span(1)
                wnd.document.mode.delete_string(
                    wnd, f, t)
            return
        else:
            tol, eol = wnd.document.mode.get_line_sel(wnd)
            wnd.screen.selection.set_range(tol, eol)

            if wnd.document.undo:
                wnd.document.undo.beginblock()
            try:
                mode = wnd.document.mode
                while tol < wnd.screen.selection.get_end():
                    m = self._is_comment_line(wnd, tol)
                    if m:
                        f, t = m.span(1)
                        wnd.document.mode.delete_string(
                            wnd, f, t, update_cursor=False)
                    tol = wnd.document.geteol(tol)
            finally:
                if wnd.document.undo:
                    wnd.document.undo.endblock()

            f, t = wnd.screen.selection.get_selrange()
            wnd.cursor.setpos(f)
            wnd.cursor.savecol()


class MacroCommands(Commands):

    @command('macro.start-record')
    @norec
    @norerun
    def start_record(self, wnd):
        kaa.app.macro.start_record()
        kaa.app.messagebar.update()

    @command('macro.end-record')
    @norec
    @norerun
    def end_record(self, wnd):
        kaa.app.macro.end_record()
        kaa.app.messagebar.update()

    @command('macro.toggle-record')
    @norec
    @norerun
    def toggle_record(self, wnd):
        kaa.app.macro.toggle_record()
        kaa.app.messagebar.update()

    @command('macro.run')
    @norec
    def run_macro(self, wnd):
        if kaa.app.macro.is_recording():
            return
        if not kaa.app.macro.get_commands():
            return

        if wnd.document.undo:
            wnd.document.undo.beginblock()
        try:
            kaa.app.macro.run(wnd)
        finally:
            if wnd.document.undo:
                wnd.document.undo.endblock()


class SearchCommands(Commands):

    @command('search.showsearch')
    @norec
    @norerun
    def showsearch(self, wnd):
        from kaa.ui.searchdlg import searchdlgmode

        buf = document.Buffer()
        doc = document.Document(buf)
        mode = searchdlgmode.SearchDlgMode(target=wnd)
        doc.setmode(mode)
        mode.build_document()

        kaa.app.show_inputline(doc)

    @command('search.showreplace')
    @norec
    @norerun
    def showreplace(self, wnd):
        from kaa.ui.searchdlg import searchdlgmode

        buf = document.Buffer()
        doc = document.Document(buf)
        mode = searchdlgmode.ReplaceDlgMode(target=wnd)
        doc.setmode(mode)
        mode.build_document()

        kaa.app.show_inputline(doc)

    def _show_searchresult(self, wnd, hit):
        if hit:
            f, t = hit.span()
            wnd.cursor.setpos(f)
            wnd.screen.selection.set_range(f, t)
            kaa.app.messagebar.set_message('found')
        else:
            kaa.app.messagebar.set_message('not found')

    @command('search.next')
    @norerun
    def searchnext(self, wnd):
        if not modebase.SearchOption.LAST_SEARCH:
            return
        if not modebase.SearchOption.LAST_SEARCH.text:
            return
        range = wnd.screen.selection.get_selrange()
        if range:
            start = range[0] + 1
        else:
            start = wnd.cursor.pos

        ret = wnd.document.mode.search_next(
            start, modebase.SearchOption.LAST_SEARCH)
        self._show_searchresult(wnd, ret)

        if not ret:
            if start != 0:
                ret = wnd.document.mode.search_next(
                    0, modebase.SearchOption.LAST_SEARCH)
                self._show_searchresult(wnd, ret)

    @command('search.prev')
    @norerun
    def searchprev(self, wnd):
        if not modebase.SearchOption.LAST_SEARCH:
            return
        if not modebase.SearchOption.LAST_SEARCH.text:
            return
        range = wnd.screen.selection.get_selrange()
        if range:
            start = range[1] - 1
        else:
            start = wnd.cursor.pos

        ret = wnd.document.mode.search_prev(
            start, modebase.SearchOption.LAST_SEARCH)
        self._show_searchresult(wnd, ret)

        if not ret:
            if start != wnd.document.endpos():
                ret = wnd.document.mode.search_prev(wnd.document.endpos(),
                                                    modebase.SearchOption.LAST_SEARCH)
                self._show_searchresult(wnd, ret)

    @command('search.showgrep')
    @norec
    @norerun
    def showgrep(self, wnd):
        from kaa.ui.grep import grepdlgmode

        buf = document.Buffer()
        doc = document.Document(buf)
        mode = grepdlgmode.GrepDlgMode(wnd)
        doc.setmode(mode)
        mode.build_document()

        kaa.app.show_dialog(doc)



from unittest.mock import patch
import kaa_testutils
from kaa import highlight
from kaa.filetype.rst import rstmode


class TestRstHighlight(kaa_testutils._TestDocBase):
    tokenizers = [rstmode.build_tokenizer()]
    tokens = tokenizers[0].tokens

    def test_inline(self):
        hl = highlight.Highlighter(tokenizers=self.tokenizers)

        doc = self._getdoc('**abc**')
        assert [
            (0, 2, self.tokens.strong.span_start),
            (2, 5, self.tokens.strong.span_mid),
            (5, 7, self.tokens.strong.span_end),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

        doc = self._getdoc('*abc*')
        assert [
            (0, 1, self.tokens.emphasis.span_start),
            (1, 4, self.tokens.emphasis.span_mid),
            (4, 5, self.tokens.emphasis.span_end),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

        doc = self._getdoc('``abc``')
        assert [
            (0, 2, self.tokens.literal.span_start),
            (2, 5, self.tokens.literal.span_mid),
            (5, 7, self.tokens.literal.span_end),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

        doc = self._getdoc('`abc`_')
        assert [
            (0, 1, self.tokens.interpreted.span_start),
            (1, 4, self.tokens.interpreted.span_mid),
            (4, 6, self.tokens.interpreted.span_end),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

        doc = self._getdoc('abc_')
        assert [
            (0, 4, self.tokens.reference.tokenid),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

        doc = self._getdoc('_`abc`')
        assert [
            (0, 3, self.tokens.target.span_start),
            (3, 5, self.tokens.target.span_mid),
            (5, 6, self.tokens.target.span_end),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

        doc = self._getdoc('|abc|')
        assert [
            (0, 5, self.tokens.substitution.tokenid),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

        doc = self._getdoc('[abc]_')
        assert [
            (0, 6, self.tokens.citation.tokenid),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

    def test_inline_delim(self):
        hl = highlight.Highlighter(tokenizers=self.tokenizers)

        doc = self._getdoc('abc*abc')
        assert [
            (0, 3, self.tokenizers[0].nulltoken),
            (3, 4, self.tokenizers[0].nulltoken),
            (4, 7, self.tokenizers[0].nulltoken),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

        doc = self._getdoc('*abc*abc')
        assert [
            (0, 1, self.tokens.emphasis.span_start),
            (1, 8, self.tokens.emphasis.span_mid),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

    def test_header(self):
        hl = highlight.Highlighter(tokenizers=self.tokenizers)

        doc = self._getdoc('--\nab\n---')
        assert [
            (0, 9, self.tokens.header1.tokenid),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

    def test_block(self):
        hl = highlight.Highlighter(tokenizers=self.tokenizers)

        doc = self._getdoc('abc:: \n')
        assert [
            (0, 3, self.tokenizers[0].nulltoken),
            (3, 6, self.tokens.block.tokenid),
            (6, 7, self.tokenizers[0].nulltoken),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

    def test_directive(self):
        hl = highlight.Highlighter(tokenizers=self.tokenizers)

        doc = self._getdoc('.. abc:: 111\n 222\n333')
        assert [
            (0,  8, self.tokens.directive.span_start),
            (8,  18, self.tokens.directive.span_mid),
            (18, 21, self.tokenizers[0].nulltoken),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

    def test_table(self):
        hl = highlight.Highlighter(tokenizers=self.tokenizers)

        with patch.object(hl, 'get_token', return_value=self.tokens.tableborder):
            doc = self._getdoc('+=+\n| |')
            assert [
                (0,  4, self.tokens.tableborder.tokenid),
                (4,  6, self.tokens.tablerow.tokenid),
                (6,  7, self.tokens.tablerow.tokenid),
            ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))




cdef public class GappedBuffer [object GappedBufferObj, type GappedBufferType]:
    cdef public Py_ssize_t bufsize
    cdef public Py_ssize_t numelems 
    cdef public Py_ssize_t gap
    cdef public Py_ssize_t gapsize
    cdef Py_UCS4 *buf

#    cdef inline _conv_slice(self, Py_ssize_t pos) nogil:
#        if pos < 0:
#
#        return self.numelems

    cdef void _resize_buf(self, Py_ssize_t size)
    cdef void _expand_gap(self, Py_ssize_t size)
    cdef void _shrink_gap(self)
    cdef void _move_gap(self, Py_ssize_t index)

    cdef Py_UCS4 _get_max(self, Py_ssize_t begin, Py_ssize_t end) nogil
    cdef void _copy_chars1(self, void *dest, Py_ssize_t begin, 
            Py_ssize_t end) nogil
    cdef void _copy_chars2(self, void *dest, Py_ssize_t begin, 
            Py_ssize_t end) nogil 
    cdef void _copy_chars4(self, void *dest, Py_ssize_t begin, 
            Py_ssize_t end) nogil
          
    cdef int _get_slice(self, object obj, 
            Py_ssize_t *f, Py_ssize_t *t) except *

    cdef _insert(self, Py_ssize_t index, unicode s)
    cpdef insert(self, Py_ssize_t index, unicode s)
    cpdef append(self, unicode s)
    cdef _delete(self, Py_ssize_t begin, Py_ssize_t end)
    cpdef delete(self, Py_ssize_t begin, Py_ssize_t end)
    cpdef replace(self, Py_ssize_t begin, Py_ssize_t end, unicode s)
    cpdef get(self, Py_ssize_t begin, Py_ssize_t end)

    cdef int _get_slice(self, object obj,
            Py_ssize_t *f, Py_ssize_t *t) except *
    cpdef getint(self, Py_ssize_t pos)
    cpdef getints(self, Py_ssize_t begin, Py_ssize_t end)
    cdef _insertints(self, Py_ssize_t index, object s)
    cpdef insertints(self, Py_ssize_t index, object s)
    cpdef appendints(self, object s)
    cpdef replaceints(self, Py_ssize_t begin, Py_ssize_t end, object s)
    cpdef setints(self, Py_ssize_t begin, Py_ssize_t end, unsigned long v)






import kaa
from kaa.command import Commands, command, norec, norerun
from kaa import document
from kaa.ui.dialog import dialogmode
from kaa.theme import Theme, Style
from kaa.keyboard import *

FrameListThemes = {
    'basic': [],
}

framelist_keys = {
    left: 'framelist.prev',
    (ctrl, 'p'): 'framelist.prev',
    (ctrl, 'b'): 'framelist.prev',
    right: 'framelist.next',
    (ctrl, 'n'): 'framelist.next',
    (ctrl, 'f'): 'framelist.next',
    '\r': 'framelist.close',
    '\n': 'framelist.close',
}


class FrameListCommands(Commands):

    def _itermarks(self, wnd):
        ret = []
        names = dict(('childframe_{}'.format(id(frame)), frame)
                     for frame in kaa.app.get_frames())
        for name, pos in wnd.document.marks.items():
            if name in names:
                ret.append((names[name], pos))
        ret = sorted(ret, key=lambda e: e[1])
        return [f for f, pos in ret]

    # todo: refactor
    @command('framelist.prev')
    @norec
    @norerun
    def prev(self, wnd):
        cur = kaa.app.get_activeframe()
        frames = self._itermarks(wnd)
        for n, frame in enumerate(frames):
            if cur is frame:
                if n:
                    frames[n - 1].bring_top()
                    wnd.get_label('popup').bring_top()
                    wnd.document.mode._update_style(wnd)
                    wnd.document.mode._activated_frame = frames[n - 1]
                    return

    @command('framelist.next')
    @norec
    @norerun
    def next(self, wnd):
        cur = kaa.app.get_activeframe()
        frames = self._itermarks(wnd)

        for n, frame in enumerate(frames):
            if cur is frame:
                if n < len(frames) - 1:
                    frames[n + 1].bring_top()
                    wnd.get_label('popup').bring_top()
                    wnd.document.mode._update_style(wnd)
                    wnd.document.mode._activated_frame = frames[n + 1]
                    return

    @command('framelist.close')
    @norec
    @norerun
    def close(self, wnd):
        # bring the frame a top of frame list.
        if wnd.document.mode._activated_frame:
            wnd.mainframe.register_childframe(
                wnd.document.mode._activated_frame)

        # Destroy popup window
        popup = wnd.get_label('popup')
        if popup:
            popup.destroy()


class FrameListMode(dialogmode.DialogMode):
    autoshrink = True
    USE_UNDO = False
    _activated_frame = None

    @classmethod
    def build(cls):
        buf = document.Buffer()
        doc = document.Document(buf)
        mode = cls()
        doc.setmode(mode)

        with dialogmode.FormBuilder(doc) as f:
            for frame in kaa.app.get_frames():
                start = f.document.endpos()
                markname = '_{}'.format(id(frame))
                f.append_text('default', frame.get_title().replace('&', '&&'))

                f.document.marks['childframe' +
                                 markname] = (start, f.document.endpos())

                f.append_text('default', ' ')

            mode._update_style(None)
            return doc

    def _update_style(self, wnd):
        activeframe = kaa.app.get_activeframe()
        cursorpos = None
        for frame in kaa.app.get_frames():
            markname = '_{}'.format(id(frame))

            f, t = self.document.marks['childframe' + markname]
            if activeframe is frame:
                self.document.setstyles(
                    f, t, self.get_styleid('activemark'))
                cursorpos = f
            else:
                self.document.setstyles(
                    f, t, self.get_styleid('nonactivemark'))

        if wnd and cursorpos is not None:
            wnd.cursor.setpos(cursorpos)

    def init_keybind(self):
        self.keybind.add_keybind(framelist_keys)

    def init_commands(self):
        super().init_commands()

        self.framelist_commands = FrameListCommands()
        self.register_command(self.framelist_commands)

    def init_themes(self):
        super().init_themes()
        self.themes.append(FrameListThemes)

    def is_cursor_visible(self):
        return 0   # hide cursor

    def on_esc_pressed(self, wnd, event):
        super().on_esc_pressed(wnd, event)
        self.framelist_commands.close(wnd)

    def on_str(self, wnd, s):
        pass



import os
import unicodedata
import sys
import locale
import kaa
import kaa.log
from kaa import document, encodingdef
from kaa.filetype.default import defaultmode
from kaa import consts
from kaa import encodingdef

class FileStorage:

    def adjust_encoding(self, encoding):
        if encoding.lower() == 'shiftjis':
            return 'cp932'
        encoding = encodingdef.normalize_encname(encoding)
        return encoding

    def guess_japanese_encoding(self, filename):
        import pyjf3
        ret = pyjf3.guess(open(filename, 'rb').read())

        if ret == pyjf3.ASCII:
            return kaa.app.config.DEFAULT_ENCODING
        elif ret == pyjf3.UTF8:
            return 'utf-8'
        elif ret == pyjf3.SJIS:
            return 'cp932'
        elif ret == pyjf3.EUC:
            return 'euc-jp'
        elif ret == pyjf3.JIS:
            return 'iso-2022-jp'
        elif ret == pyjf3.UTF16_LE:
            return 'utf-16le'
        elif ret == pyjf3.UTF16_BE:
            return 'utf-16be'

    def get_textio(self, fileinfo, filemustexists=False):
        try:
            # use surrogateescape to preserve file contents intact.
            textio = open(fileinfo.fullpathname, 'r',
                          encoding=self.adjust_encoding(fileinfo.encoding),
                          errors='surrogateescape',
                          newline=fileinfo.nlchars)

        except FileNotFoundError:
            if filemustexists:
                raise
            return None

        return textio

    def get_fileinfo(self, filename, encoding=None, newline=None):
        if sys.platform == 'darwin':
            filename = unicodedata.normalize('NFC', filename)

        if encoding is None:
            encoding = kaa.app.config.DEFAULT_ENCODING

        if newline is None:
            newline = kaa.app.config.DEFAULT_NEWLINE

        fileinfo = FileInfo()

        fullpath = os.path.abspath(filename)
        dirname, filename = os.path.split(filename)

        fileinfo.storage = self
        fileinfo.fullpathname = fullpath
        fileinfo.dirname = dirname
        fileinfo.filename = filename

        try:
            fileinfo.stat = os.stat(fullpath)
        except FileNotFoundError:
            fileinfo.stat = None

        fileinfo.encoding = encoding
        fileinfo.newline = newline

        return fileinfo

    def listdir(self, dirname):
        dirs = []
        files = []
        for name in os.listdir(dirname):
            if os.path.isdir(os.path.join(dirname, name)):
                dirs.append(name)
            else:
                files.append(name)
        if sys.platform == 'darwin':
            dirs = [unicodedata.normalize('NFC', n) for n in dirs]
            files = [unicodedata.normalize('NFC', n) for n in files]
        return dirs, files

    def _save_hist(self, filename):
        abspath = os.path.abspath(filename)
        dirname = os.path.dirname(abspath)
        if dirname.endswith(os.pathsep):
            dirname = dirname[:-1]

        kaa.app.config.hist('filename').add(abspath)
        kaa.app.config.hist('dirname').add(dirname)
        kaa.app.last_dir = dirname

    def newfile(self, mode=None, s='', temporary=False):
        buf = document.Buffer()
        if s:
            buf.insert(0, s)

        doc = document.Document(buf)
        doc.temporary = temporary

        doc.title = '<Untitled {}>'.format(kaa.app.NUM_NEWFILE)
        kaa.app.NUM_NEWFILE += 1

        doc.fileinfo = FileInfo()
        if not mode:
            mode = defaultmode.DefaultMode()
        doc.setmode(mode)
        return doc

    def openfile(self, filename, encoding=None, newline=None, nohist=False,
                 filemustexists=False):
        if not nohist:
            self._save_hist(filename)

        fileinfo = self.get_fileinfo(filename, encoding, newline)
        modecls = select_mode(fileinfo.fullpathname)
        modecls.update_fileinfo(fileinfo)
        
        textio = self.get_textio(fileinfo, filemustexists)

        buf = document.Buffer()
        if textio:
            src = textio.read()
            if fileinfo.nlchars is not None:
                src = src.replace(fileinfo.nlchars, '\n')
            buf.insert(0, src)

        doc = document.Document(buf)
        doc.fileinfo = fileinfo

        doc.setmode(modecls())

        dir, file = os.path.split(fileinfo.fullpathname)
        if not dir.endswith(os.path.sep):
            dir += os.path.sep

        kaa.app.messagebar.set_message('Read from {}({})'.format(file, dir))

        kaa.app.config.hist_storage.flush()
        return doc


    def save_document(self, doc, filename, encoding=None, newline=None,
                      nohist=False):
        # todo: move most of codes to mode class.
        if not nohist:
            self._save_hist(filename)

        if not doc.fileinfo:
            fileinfo = kaa.app.storage.get_fileinfo(
                filename, encoding, newline)
            doc.fileinfo = fileinfo
        else:
            if encoding is not None:
                doc.fileinfo.encoding = encoding

            if doc.fileinfo.encoding is None:
                doc.fileinfo.encoding = kaa.app.config.DEFAULT_ENCODING

            if newline is not None:
                doc.fileinfo.newline = newline

            if doc.fileinfo.newline is None:
                doc.fileinfo.newline = kaa.app.config.DEFAULT_NEWLINE

        modecls = select_mode(filename)
        if type(doc.mode) is not modecls:
            doc.setmode(modecls())

        doc.mode.update_fileinfo(doc.fileinfo, doc)

        with open(filename, 'w',
                  encoding=self.adjust_encoding(doc.fileinfo.encoding),
                  newline=consts.NEWLINE_CHARS[doc.fileinfo.newline],
                  errors='surrogateescape') as f:

            # TODO: save as another file and rename.
            # TODO: fsync
            f.write(doc.gettext(0, doc.endpos()))

        fileinfo = self.get_fileinfo(filename, doc.fileinfo.encoding,
                                     doc.fileinfo.newline)
        doc.mode.on_file_saved(fileinfo)

        doc.fileinfo = fileinfo
        doc.fileinfo.check_update() # ensure timestamps updated

        doc.set_title(None)
        doc.temporary = False
        if doc.undo:
            doc.undo.saved()

        dir, file = os.path.split(fileinfo.fullpathname)
        if not dir.endswith(os.path.sep):
            dir += os.path.sep

        kaa.app.messagebar.set_message('Written to {}({})'.format(file, dir))

        kaa.app.config.hist_storage.flush()

class FileInfo:
    storage = None
    fullpathname = ''
    dirname = ''
    filename = ''
    stat = None
    encoding = None
    newline = None

    @property
    def nlchars(self):
        return consts.NEWLINE_CHARS[self.newline]

    def __init__(self):
        self.encoding = kaa.app.config.DEFAULT_ENCODING
        self.newline = kaa.app.config.DEFAULT_NEWLINE

    def save_screenloc(self, pos):
        if self.fullpathname:
            info = kaa.app.config.hist('filedisp').find(self.fullpathname)
            if not info:
                info = {}
            info['pos'] = pos
            kaa.app.config.hist('filedisp').add(
                self.fullpathname, info)

    def check_update(self):
        if self.fullpathname:
            try:
                stat = os.stat(self.fullpathname)
            except Exception:
                return

            if stat.st_mtime == self.stat.st_mtime:
                return

            if stat.st_size == self.stat.st_size:
                return

            self.stat = stat
            return True


def select_mode(filename):
    for pkg in kaa.app.config.get_mode_packages():
        if hasattr(pkg, 'FileTypeInfo'):
            ret = pkg.FileTypeInfo.select_mode(filename)
            if ret:
                return ret

    return defaultmode.DefaultMode


#def openfile(filename, encoding=None, newline=None, filemustexists=False):
#    # Open file
#    fileinfo = kaa.app.storage.get_fileinfo(filename, encoding, newline)
#    textio = kaa.app.storage.get_textio(fileinfo, filemustexists)
#
#    buf = document.Buffer()
#    if textio:
#        src = textio.read()
#        if fileinfo.nlchars is not None:
#            src = src.replace(fileinfo.nlchars, '\n')
#        buf.insert(0, src)
#
#    doc = document.Document(buf)
#    doc.fileinfo = fileinfo
#    doc.setmode(select_mode(filename)())
#
#    dir, file = os.path.split(fileinfo.fullpathname)
#    if not dir.endswith(os.path.sep):
#        dir += os.path.sep
#
#    kaa.app.messagebar.set_message('Read from {}({})'.format(file, dir))
#
#    return doc
#

#
#def newfile(mode=None, s='', provisional=False):
#    buf = document.Buffer()
#    if s:
#        buf.insert(0, s)
#
#    doc = document.Document(buf)
#    doc.provisional = provisional
#
#    global NUM_NEWFILE
#    doc.title = '<Untitled {}>'.format(NUM_NEWFILE)
#    NUM_NEWFILE += 1
#
#    doc.fileinfo = FileInfo()
#    if not mode:
#        mode = defaultmode.DefaultMode()
#    doc.setmode(mode)
#    return doc



import unicodedata
import copy
from collections import namedtuple
from kaa.filetype.default import defaultmode
from gappedbuf import re as gre
from kaa.highlight import Tokenizer, Token, Span, Keywords, EndSection, SingleToken
from kaa.theme import Theme, Style
from kaa.command import Commands, command, norec, norerun
from kaa.keyboard import *

RstThemes = {
    'basic': [
        Style('header', 'Blue', None),
        Style('block', 'Orange', None),
        Style('directive', 'Green', None),
        Style('table', 'Cyan', None),
        Style('strong', 'Magenta', None),
        Style('emphasis', 'Blue', None),
        Style('literal', 'Cyan', None),
        Style('reference', 'Red', None),
        Style('role', 'Cyan', None),
        Style('substitution', 'Green', None),
    ],
}


def _build_seps():
    starts = []
    ends = []
    
    for i in range(0x12000):
        c = chr(i)
        category = unicodedata.category(c)
        if category in {'Pd', 'Po', 'Ps', 'Pi', 'Pf'}:
            starts.append(c)
        if category in {'Pd', 'Po', 'Pe', 'Pi', 'Pf'}:
            ends.append(c)
    return ''.join(starts), ''.join(ends)

SEP_STARTS, SEP_ENDS = _build_seps()

class RstInline(Span):
    WS = ' \t\r\n'
    STARTS = '\'"([{<' + WS + SEP_STARTS
    ENDS = '\'".,:;!?-)]}/\\>' + WS + SEP_ENDS

    def on_start(self, tokenizer, doc, pos, match):
        if ((pos and (doc.gettext(pos - 1, pos) not in self.STARTS)) or
                (pos >= doc.endpos() - 1) or
                (doc.gettext(pos + 1, pos + 2) in self.WS)):

            yield pos, pos + 1, tokenizer.nulltoken
            return pos + 1, None, False

        ret = yield from super().on_start(tokenizer, doc, pos, match)
        return ret

    def _is_end(self, doc, m):
        pos = m.end()
        if pos < doc.endpos():
            return (doc.gettext(pos, pos + 1) in self.ENDS)
        return True


class TableToken(SingleToken):

    def on_start(self, tokenizer, doc, pos, match):
        in_table = False
        tol = doc.gettol(pos)
        if pos == tol:
            in_table = True
        else:
            # check if previous line is table
            if tol > 1:
                laststyle = doc.getstyles(tol - 2, tol - 1)[0]
                token = tokenizer.highlighter.get_token(laststyle)
                if isinstance(token, TableToken):
                    in_table = True
        if in_table:
            ret = yield from super().on_start(tokenizer, doc, pos, match)
            return ret
        else:
            yield pos, pos + 1, tokenizer.nulltoken
            return pos + 1, None, False


RST_HEADERS = r'=\-`:\'"~^_*+#<>'


def build_tokenizer():
    RSTTOKENS = namedtuple('rsttokens',
                           ['escape', 'header1', 'header2',
                            'directive', 'block', 'tableborder',
                            'tablerow',
                            'strong', 'emphasis',
                            'literal', 'interpreted', 'reference',
                            'role',
                            'target', 'substitution', 'citation',
                            ])

    return Tokenizer(
        RSTTOKENS(
            # escape
            SingleToken('escape', 'default', [r'\\.']),

            # header token
            SingleToken('header1', 'header',
                        [r'^(?P<H>[{}])(?P=H)+\n.+\n(?P=H)+$'.format(RST_HEADERS)]),
            SingleToken('header2', 'header',
                        [r'^.+\n(?P<H2>[{}])(?P=H2)+$'.format(RST_HEADERS)]),

            # block
            Span('directive', 'directive', r'\.\.\s+\S+::', '^\S',
                 escape='\\', capture_end=False),
            SingleToken('block', 'block', [r'::[\ \t]*$']),

            # table
            TableToken('rst-table-border', 'table', [r'\+[+\-=]+(\s+|$)']),
            TableToken('rst-table-row', 'table', [r'\|(\s+|$)']),

            # inline token
            RstInline('rst-strong', 'strong',
                      r'\*\*', r'\*\*', escape='\\'),
            RstInline('rst-emphasis', 'emphasis',
                      r'\*', r'\*', escape='\\'),
            RstInline('rst-literal', 'literal', r'``', r'``', escape='\\'),
            RstInline('rst-interpreted', 'reference', r'`', r'`_?_?',
                      escape='\\'),
            SingleToken('rst-reference', 'reference',
                        [r'\b[a-zA-Z0-9]+__?\b']),
            RstInline('rst-role', 'role', r':[a-zA-Z0-9]+:`', r'`',
                      escape='\\'),
            RstInline('rst-target', 'reference',
                      r'_`\w', r'`', escape='\\'),
            SingleToken('rst-substitution', 'substitution', [r'\|\w+\|']),
            SingleToken('rst-citation', 'reference', [r'\[\w+\]_\b']),

        ))


RSTMENU = [
    ['&Table of contents', None, 'toc.showlist'],
]

rst_keys = {
    (ctrl, 't'): 'toc.showlist',
}


class RstMode(defaultmode.DefaultMode):
    MODENAME = 'Rst'

    def init_keybind(self):
        super().init_keybind()
        self.register_keys(self.keybind, [rst_keys])

    def init_menu(self):
        super().init_menu()
        self.menu['CODE'] = copy.deepcopy(RSTMENU)

    def init_themes(self):
        super().init_themes()
        self.themes.append(RstThemes)

    def init_tokenizers(self):
        self.tokenizers = [build_tokenizer()]

    HEADER1 = r'''^(?P<H>[{}])(?P=H)+\n
                  (?P<TITLE>.+)\n
                  (?P=H)+$'''.format(RST_HEADERS)

    HEADER2 = r'''^(?P<TITLE2>.+)\n
                (?P<H2>[{}])(?P=H2)+$'''.format(RST_HEADERS)

    RE_HEADER = gre.compile('|'.join([HEADER1, HEADER2]), gre.X | gre.M)

    def get_headers(self):
        levels = {}
        stack = []
        pos = 0
        while True:
            m = self.RE_HEADER.search(self.document.buf, pos)
            if not m:
                break

            pos = m.end()
            name = m.group('TITLE') or m.group('TITLE2')
            name = name.strip()

            bar = (m.group('H'), m.group('H2'))
            if bar not in levels:
                level = len(levels)
                levels[bar] = level
            else:
                level = levels[bar]

            if not stack:
                header = self.HeaderInfo('namespace', None, name, name,
                                         None, m.start())
                yield header
                stack.append((level, header))
                continue

            for i, (parent_level, info) in enumerate(stack):
                if level <= parent_level:
                    del stack[i:]
                    break

            if stack:
                parent = stack[-1][1]
            else:
                parent = None

            header = self.HeaderInfo(
                'namespace', parent, name,
                name, None, m.start())
            yield header
            stack.append((level, header))

    @command('toc.showlist')
    @norec
    @norerun
    def show_toclist(self, wnd):
        from kaa.ui.toclist import toclistmode
        headers = list(self.get_headers())
        toclistmode.show_toclist(wnd, headers)



import curses
import collections
import kaa
from kaa import document
from kaa.cui.color import ColorName
from kaa.ui.statusbar import statusbarmode
from kaa.cui import editor


class Splitter:
    vsep = hsep = 0

    def __init__(self, frame, parent, wnd):
        super().__init__()

        self.wnd = self.left = self.right = self.above = self.below = None
        assert frame
        self.frame = frame
        self.parent = parent
        self.wnd = wnd
        self.wnd.set_splitter(self)
        self.statusbar = None
        self.rect = (0, 0, 0, 0)

    def walk(self):
        stack = collections.deque([self])
        while stack:
            v = stack.pop()
            yield (v, v.wnd)
            if v.left:
                stack.append(v.right)
                stack.append(v.left)
            elif v.above:
                stack.append(v.below)
                stack.append(v.above)

    def activate(self):
        if self.wnd:
            self.wnd.activate()
        elif self.left:
            self.left.activate()
            self.right.activate()
        elif self.above:
            self.above.activate()
            self.below.activate()

    def set_rect(self, l, t, r, b):
        self.rect = (l, t, r, b)
        self.resize()

    def _destroy_children(self):
        if self.left:
            self.left.destroy()
        if self.right:
            self.right.destroy()
        if self.above:
            self.above.destroy()
        if self.below:
            self.below.destroy()
        self.left = self.right = self.above = self.below = None

    def destroy(self):
        self._destroy_children()
        if self.wnd:
            self.wnd.destroy()
        if self.statusbar:
            self.statusbar.destroy()

        self.parent = None
        self.frame = None
        self.statusbar = None

    def get_buddy(self):
        if self.parent:
            if self.parent.left is self:
                return self.parent.right
            elif self.parent.right is self:
                return self.parent.left
            elif self.parent.above is self:
                return self.parent.below
            elif self.parent.below is self:
                return self.parent.above

    def show_doc(self, doc):
        if not self.wnd:
            self._destroy_children()
            self.wnd = editor.TextEditorWindow(parent=self.frame)
            self.resize()

        self.wnd.show_doc(doc)

    def split(self, vert, doc=None):
        w, h = self.wnd.getsize()
        if not doc:
            doc = self.wnd.document
        if vert:
            self.vsep = w // 2 + 1
            self.left = Splitter(self.frame, self, self.wnd)
            self.right = Splitter(self.frame, self, self.wnd.dup())
            self.right.wnd.show_doc(doc)
            self.wnd = None
            self.left.activate()
            ret = self.right
        else:
            self.hsep = h // 2 + 1
            self.above = Splitter(self.frame, self, self.wnd)
            self.below = Splitter(self.frame, self, self.wnd.dup())
            self.below.wnd.show_doc(doc)
            self.wnd = None
            self.above.activate()
            ret = self.below

        self.resize()
        return ret

    def join(self, wnd):
        if self.wnd:
            # not splitted
            return

        next = None
        for w in (self.left, self.right, self.above, self.below):
            if w:
                if wnd is next is None:
                    # Keep first window if wnd is None
                    next = w
                elif wnd is not w.wnd:
                    w.destroy()
                else:
                    next = w

        # chain child splitter to parent
        next.parent = self.parent
        next.rect = self.rect
        if self.frame.splitter is self:
            self.frame.splitter = next
        else:
            if self.parent.left is self:
                self.parent.left = next
            elif self.parent.right is self:
                self.parent.right = next
            elif self.parent.above is self:
                self.parent.above = next
            elif self.parent.below is self:
                self.parent.below = next

        # clean up
        self.left = self.right = self.above = self.below = None
        self.destroy()

        next.resize()
        next.activate()

    def _build_statusbar(self):
        self.statusbar = editor.TextEditorWindow(parent=self.frame)

        buf = document.Buffer()
        doc = document.Document(buf)
        mode = statusbarmode.StatusBarMode()
        doc.setmode(mode)
        self.statusbar.show_doc(doc)

    def resize(self):
        l, t, r, b = self.rect
        if self.wnd:
            if not self.statusbar:
                self._build_statusbar()
            self.statusbar.set_rect(l, b - 1, r, b)
            self.statusbar.bring_top()
            self.wnd.set_rect(l, t, r, b - 1)
            self.wnd.set_statusbar(self.statusbar.document.mode)
            self.wnd.bring_top()
        else:
            if self.statusbar:
                self.statusbar.destroy()
                self.statusbar = None

            if self.left:
                self.left.set_rect(l, t, l + self.vsep, b)
                self.right.set_rect(l + self.vsep + 1, t, r, b)
            elif self.above:
                self.above.set_rect(l, t, r, t + self.hsep)
                self.below.set_rect(l, t + self.hsep, r, b)

    def get_editors(self):
        if self.wnd:
            return [self.wnd]
        elif self.left:
            return self.left.get_editors() + self.right.get_editors()
        else:
            return self.above.get_editors() + self.below.get_editors()

    def get_title(self):
        if self.wnd:
            doc = self.wnd.document
            return [doc.get_title()]

        elif self.left:
            return self.left.get_title() + self.right.get_title()

        else:
            return self.above.get_title() + self.below.get_title()

    def bring_top(self):
        if self.wnd:
            if self.statusbar:
                self.statusbar.bring_top()
            self.wnd.bring_top()

        elif self.left:
            self.left.bring_top()
            self.right.bring_top()

        else:
            self.above.bring_top()
            self.below.bring_top()

    def draw(self):
        self.frame._cwnd.attron(
            kaa.app.colors.get_color(
                kaa.app.colors.colornames.get('CYAN'),
                kaa.app.colors.colornames.get('DEFAULT')))
        l, t, r, b = self.rect
        if self.left:
            try:
                self.frame._cwnd.vline(
                    t, l + self.vsep, curses.ACS_VLINE, b - t)
            except curses.error:
                pass

            self.left.draw()
            self.right.draw()
        elif self.above:
            self.above.draw()
            self.below.draw()

    def separator_prev(self):
        if self.left:
            if self.vsep > 3:
                self.vsep -= 1
                self.resize()
                self.draw()
        elif self.above:
            if self.hsep > 3:
                self.hsep -= 1
                self.resize()
                self.draw()

    def separator_next(self):
        l, t, r, b = self.rect

        if self.left:
            if r - (l + self.vsep) > 4:
                self.vsep += 1
                self.resize()
                self.draw()
        elif self.above:
            if b - (t + self.hsep) > 3:
                self.hsep += 1
                self.resize()
                self.draw()




cdef extern from "Python.h":
    ctypedef unsigned char Py_UCS1
    ctypedef unsigned short Py_UCS2

    Py_UCS4* PyUnicode_AsUCS4Copy(object u) except NULL
    Py_UCS4* PyUnicode_AsUCS4(object u, Py_UCS4 *buffer, 
            Py_ssize_t buflen, int copy_null) except NULL
    object PyUnicode_New(Py_ssize_t size, Py_UCS4 maxchar)
    void* PyUnicode_DATA(object o) except NULL
    int PyUnicode_KIND(object o) except 0

    int PySlice_GetIndicesEx(object slice, Py_ssize_t length,
            Py_ssize_t *start, Py_ssize_t *stop, Py_ssize_t *step, 
            Py_ssize_t *slicelength) except -1
    int PySlice_Check(object)




from kaa.filetype import filetypedef

class FileTypeInfo(filetypedef.FileTypeInfo):
    FILE_EXT = {}

    @classmethod
    def get_modetype(cls):
        from kaa.filetype.default.defaultmode import DefaultMode
        return DefaultMode



import curses


class ColorName:
    DEFAULT = -1
    BLACK = curses.COLOR_BLACK
    BLUE = curses.COLOR_BLUE
    CYAN = curses.COLOR_CYAN
    GREEN = curses.COLOR_GREEN
    MAGENTA = curses.COLOR_MAGENTA
    RED = curses.COLOR_RED
    WHITE = curses.COLOR_WHITE
    YELLOW = curses.COLOR_YELLOW

    COLOR0 = 0
    COLOR1 = 1
    COLOR2 = 2
    COLOR3 = 3
    COLOR4 = 4
    COLOR5 = 5
    COLOR6 = 6
    COLOR7 = 7
    COLOR8 = 8
    COLOR9 = 9
    COLOR10 = 10
    COLOR11 = 11
    COLOR12 = 12
    COLOR13 = 13
    COLOR14 = 14
    COLOR15 = 15
    COLOR16 = 16
    COLOR17 = 17
    COLOR18 = 18
    COLOR19 = 19
    COLOR20 = 20
    COLOR21 = 21
    COLOR22 = 22
    COLOR23 = 23
    COLOR24 = 24
    COLOR25 = 25
    COLOR26 = 26
    COLOR27 = 27
    COLOR28 = 28
    COLOR29 = 29
    COLOR30 = 30
    COLOR31 = 31
    COLOR32 = 32
    COLOR33 = 33
    COLOR34 = 34
    COLOR35 = 35
    COLOR36 = 36
    COLOR37 = 37
    COLOR38 = 38
    COLOR39 = 39
    COLOR40 = 40
    COLOR41 = 41
    COLOR42 = 42
    COLOR43 = 43
    COLOR44 = 44
    COLOR45 = 45
    COLOR46 = 46
    COLOR47 = 47
    COLOR48 = 48
    COLOR49 = 49
    COLOR50 = 50
    COLOR51 = 51
    COLOR52 = 52
    COLOR53 = 53
    COLOR54 = 54
    COLOR55 = 55
    COLOR56 = 56
    COLOR57 = 57
    COLOR58 = 58
    COLOR59 = 59
    COLOR60 = 60
    COLOR61 = 61
    COLOR62 = 62
    COLOR63 = 63
    COLOR64 = 64
    COLOR65 = 65
    COLOR66 = 66
    COLOR67 = 67
    COLOR68 = 68
    COLOR69 = 69
    COLOR70 = 70
    COLOR71 = 71
    COLOR72 = 72
    COLOR73 = 73
    COLOR74 = 74
    COLOR75 = 75
    COLOR76 = 76
    COLOR77 = 77
    COLOR78 = 78
    COLOR79 = 79
    COLOR80 = 80
    COLOR81 = 81
    COLOR82 = 82
    COLOR83 = 83
    COLOR84 = 84
    COLOR85 = 85
    COLOR86 = 86
    COLOR87 = 87
    COLOR88 = 88
    COLOR89 = 89
    COLOR90 = 90
    COLOR91 = 91
    COLOR92 = 92
    COLOR93 = 93
    COLOR94 = 94
    COLOR95 = 95
    COLOR96 = 96
    COLOR97 = 97
    COLOR98 = 98
    COLOR99 = 99
    COLOR100 = 100
    COLOR101 = 101
    COLOR102 = 102
    COLOR103 = 103
    COLOR104 = 104
    COLOR105 = 105
    COLOR106 = 106
    COLOR107 = 107
    COLOR108 = 108
    COLOR109 = 109
    COLOR110 = 110
    COLOR111 = 111
    COLOR112 = 112
    COLOR113 = 113
    COLOR114 = 114
    COLOR115 = 115
    COLOR116 = 116
    COLOR117 = 117
    COLOR118 = 118
    COLOR119 = 119
    COLOR120 = 120
    COLOR121 = 121
    COLOR122 = 122
    COLOR123 = 123
    COLOR124 = 124
    COLOR125 = 125
    COLOR126 = 126
    COLOR127 = 127
    COLOR128 = 128
    COLOR129 = 129
    COLOR130 = 130
    COLOR131 = 131
    COLOR132 = 132
    COLOR133 = 133
    COLOR134 = 134
    COLOR135 = 135
    COLOR136 = 136
    COLOR137 = 137
    COLOR138 = 138
    COLOR139 = 139
    COLOR140 = 140
    COLOR141 = 141
    COLOR142 = 142
    COLOR143 = 143
    COLOR144 = 144
    COLOR145 = 145
    COLOR146 = 146
    COLOR147 = 147
    COLOR148 = 148
    COLOR149 = 149
    COLOR150 = 150
    COLOR151 = 151
    COLOR152 = 152
    COLOR153 = 153
    COLOR154 = 154
    COLOR155 = 155
    COLOR156 = 156
    COLOR157 = 157
    COLOR158 = 158
    COLOR159 = 159
    COLOR160 = 160
    COLOR161 = 161
    COLOR162 = 162
    COLOR163 = 163
    COLOR164 = 164
    COLOR165 = 165
    COLOR166 = 166
    COLOR167 = 167
    COLOR168 = 168
    COLOR169 = 169
    COLOR170 = 170
    COLOR171 = 171
    COLOR172 = 172
    COLOR173 = 173
    COLOR174 = 174
    COLOR175 = 175
    COLOR176 = 176
    COLOR177 = 177
    COLOR178 = 178
    COLOR179 = 179
    COLOR180 = 180
    COLOR181 = 181
    COLOR182 = 182
    COLOR183 = 183
    COLOR184 = 184
    COLOR185 = 185
    COLOR186 = 186
    COLOR187 = 187
    COLOR188 = 188
    COLOR189 = 189
    COLOR190 = 190
    COLOR191 = 191
    COLOR192 = 192
    COLOR193 = 193
    COLOR194 = 194
    COLOR195 = 195
    COLOR196 = 196
    COLOR197 = 197
    COLOR198 = 198
    COLOR199 = 199
    COLOR200 = 200
    COLOR201 = 201
    COLOR202 = 202
    COLOR203 = 203
    COLOR204 = 204
    COLOR205 = 205
    COLOR206 = 206
    COLOR207 = 207
    COLOR208 = 208
    COLOR209 = 209
    COLOR210 = 210
    COLOR211 = 211
    COLOR212 = 212
    COLOR213 = 213
    COLOR214 = 214
    COLOR215 = 215
    COLOR216 = 216
    COLOR217 = 217
    COLOR218 = 218
    COLOR219 = 219
    COLOR220 = 220
    COLOR221 = 221
    COLOR222 = 222
    COLOR223 = 223
    COLOR224 = 224
    COLOR225 = 225
    COLOR226 = 226
    COLOR227 = 227
    COLOR228 = 228
    COLOR229 = 229
    COLOR230 = 230
    COLOR231 = 231
    COLOR232 = 232
    COLOR233 = 233
    COLOR234 = 234
    COLOR235 = 235
    COLOR236 = 236
    COLOR237 = 237
    COLOR238 = 238
    COLOR239 = 239
    COLOR240 = 240
    COLOR241 = 241
    COLOR242 = 242
    COLOR243 = 243
    COLOR244 = 244
    COLOR245 = 245
    COLOR246 = 246
    COLOR247 = 247
    COLOR248 = 248
    COLOR249 = 249
    COLOR250 = 250
    COLOR251 = 251
    COLOR252 = 252
    COLOR253 = 253
    COLOR254 = 254
    COLOR255 = 255

    # from http://ethanschoonover.com/solarized
    BASE03 = 234
    BASE02 = 235
    BASE01 = 240
    BASE00 = 241
    BASE0 = 244
    BASE1 = 245
    BASE2 = 254
    BASE3 = 230
    YELLOW = 136
    ORANGE = 166
    RED = 160
    MAGENTA = 125
    VIOLET = 61
    BLUE = 33
    CYAN = 37
    GREEN = 64

    # from tmux color.c
    _COLOR256_TO_16 = [
        0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15,
        0, 4, 4, 4, 12, 12, 2, 6, 4, 4, 12, 12, 2, 2, 6, 4,
        12, 12, 2, 2, 2, 6, 12, 12, 10, 10, 10, 10, 14, 12, 10, 10,
        10, 10, 10, 14, 1, 5, 4, 4, 12, 12, 3, 8, 4, 4, 12, 12,
        2, 2, 6, 4, 12, 12, 2, 2, 2, 6, 12, 12, 10, 10, 10, 10,
        14, 12, 10, 10, 10, 10, 10, 14, 1, 1, 5, 4, 12, 12, 1, 1,
        5, 4, 12, 12, 3, 3, 8, 4, 12, 12, 2, 2, 2, 6, 12, 12,
        10, 10, 10, 10, 14, 12, 10, 10, 10, 10, 10, 14, 1, 1, 1, 5,
        12, 12, 1, 1, 1, 5, 12, 12, 1, 1, 1, 5, 12, 12, 3, 3,
        3, 7, 12, 12, 10, 10, 10, 10, 14, 12, 10, 10, 10, 10, 10, 14,
        9, 9, 9, 9, 13, 12, 9, 9, 9, 9, 13, 12, 9, 9, 9, 9,
        13, 12, 9, 9, 9, 9, 13, 12, 11, 11, 11, 11, 7, 12, 10, 10,
        10, 10, 10, 14, 9, 9, 9, 9, 9, 13, 9, 9, 9, 9, 9, 13,
        9, 9, 9, 9, 9, 13, 9, 9, 9, 9, 9, 13, 9, 9, 9, 9,
        9, 13, 11, 11, 11, 11, 11, 15, 0, 0, 0, 0, 0, 0, 8, 8,
        8, 8, 8, 8, 7, 7, 7, 7, 7, 7, 15, 15, 15, 15, 15, 15
    ]

    def __init__(self):
        self.COLOR256_TO_16 = self._COLOR256_TO_16[:]
        try:
            self.num_colors = curses.tigetnum('colors')
        except curses.error:
            self.num_colors = 8

    def get(self, name):
        ret = getattr(self, name.upper())
        if ret < 0:   # name is 'default'
            return ret
        if self.num_colors < 256:
            ret = self.COLOR256_TO_16[ret]
        if self.num_colors < 16:
            ret = ret - 8 if ret >= 8 else ret
        return ret


class DarkPalette(ColorName):
    # from http://ethanschoonover.com/solarized
    BASE03 = 234
    BASE02 = 235
    BASE01 = 240
    BASE00 = 241
    BASE0 = 244
    BASE1 = 245
    BASE2 = 254
    BASE3 = 230

    LIGHTBLUE = ColorName.BLUE

    def __init__(self):
        super().__init__()

        self.COLOR256_TO_16[self.BASE02] = curses.COLOR_BLUE
        self.COLOR256_TO_16[self.BASE3] = curses.COLOR_WHITE
        self.COLOR256_TO_16[self.YELLOW] = curses.COLOR_YELLOW
        self.COLOR256_TO_16[self.WHITE] = curses.COLOR_WHITE

        if self.num_colors < 256:
            # todo:
            self.DEFAULT = -1
            self.BLACK = curses.COLOR_BLACK
            self.BLUE = curses.COLOR_BLUE
            self.CYAN = curses.COLOR_CYAN
            self.GREEN = curses.COLOR_GREEN
            self.MAGENTA = curses.COLOR_MAGENTA
            self.RED = curses.COLOR_RED
            self.WHITE = curses.COLOR_WHITE
            self.YELLOW = curses.COLOR_YELLOW

            self.ORANGE = curses.COLOR_MAGENTA
            self.LIGHTBLUE = curses.COLOR_CYAN


class LightPalette(ColorName):
    # from http://ethanschoonover.com/solarized
    BASE3 = 234
    BASE2 = 235
    BASE1 = 240
    BASE0 = 241
    BASE00 = 244
    BASE01 = 245
    BASE02 = 254
    BASE03 = 230

    LIGHTBLUE = ColorName.CYAN

    def __init__(self):
        super().__init__()

        self.COLOR256_TO_16[self.BASE02] = curses.COLOR_BLUE
        self.COLOR256_TO_16[self.BASE3] = curses.COLOR_CYAN
        self.COLOR256_TO_16[self.YELLOW] = curses.COLOR_YELLOW
        self.COLOR256_TO_16[self.WHITE] = curses.COLOR_WHITE

        if self.num_colors < 256:
            # todo:
            self.DEFAULT = -1
            self.BLACK = curses.COLOR_BLACK
            self.BLUE = curses.COLOR_BLUE
            self.CYAN = curses.COLOR_CYAN
            self.GREEN = curses.COLOR_GREEN
            self.MAGENTA = curses.COLOR_MAGENTA
            self.RED = curses.COLOR_RED
            self.WHITE = curses.COLOR_WHITE
            self.YELLOW = curses.COLOR_YELLOW

            self.ORANGE = curses.COLOR_MAGENTA
            self.LIGHTBLUE = curses.COLOR_CYAN


class Colors:

    def __init__(self, palette):
        self.colornames = palette

        self.pairs = {
            (ColorName.WHITE, ColorName.BLACK): 0
        }

    def get_color(self, fg, bg):
        if (fg, bg) not in self.pairs:
            n = len(self.pairs)
            curses.init_pair(n, fg, bg)
            self.pairs[(fg, bg)] = curses.color_pair(n)

        return self.pairs[(fg, bg)]



import kaa.log
import cProfile
import pstats
import io
import sys
import contextlib
import time
import traceback
import pprint
import inspect


def _profile(f, *args, **kwargs):
    prof = cProfile.Profile()
    ret = prof.runcall(f, *args, **kwargs)

    out = io.StringIO()
    pstats.Stats(prof, stream=out).strip_dirs().sort_stats(
        'cumulative').print_stats()
    _trace('\n' + out.getvalue())

    return ret


def _get_caller(n=2):
    rec = inspect.getouterframes(inspect.currentframe())[n]
    frame, fname, lno, funcname, lines, index = rec

    return "{}:{} {} - {}".format(fname, lno, funcname, lines[0])


def _get_stack(limit=None):
    out = io.StringIO()
    traceback.print_stack(sys._getframe(2), limit=limit, file=out)
    _trace(out.getvalue())


def _print_exc():
    out = io.StringIO()
    traceback.print_exc(file=out)
    _trace(out.getvalue())


def _trace(*args):
    kaa.log.debug(' '.join(str(o) for o in args))


@contextlib.contextmanager
def _stime(header='--------------'):
    """
    with _stime('testing') as o:
        do_something()
        o.t1
        do_something()
        o.t2
    """
    _trace(_get_caller(3), header)

    f = time.clock()

    class ll:

        def __init__(self):
            self.values = []
            self.start = time.clock()

        def __getattr__(self, name):
            self.L(name)

        def L(self, name):
            self.values.append((name, time.clock()))
            if len(self.values) == 1:
                _trace(
                    ' - {}: {:.6}'.format(o.values[0][0], o.values[0][1] - o.start))
            else:
                t1, t2 = o.values[-2:]
                _trace('{} - {}: {:.6}'.format(t1[0], t2[0], t2[1] - t1[1]))

    o = ll()
    yield o
    o.fin = time.clock()

    if o.values:
        _trace('{} - : {:.6}'.format(o.values[-1][0], o.fin - o.values[-1][1]))

    _trace('total: {:.6}'.format(o.fin - o.start))


def _pprint(obj):
    out = io.StringIO()
    pprint.pprint(obj, stream=out)
    _trace(out.getvalue())


import builtins
builtins._profile = _profile
builtins._get_caller = _get_caller
builtins._get_stack = _get_stack
builtins._trace = _trace
builtins._stime = _stime
builtins._pprint = _pprint
builtins._print_exc = _print_exc



from collections import namedtuple
from kaa.filetype.default import defaultmode
from gappedbuf import re as gre
from kaa.highlight import (Tokenizer, Span, SingleToken, Token,
                           SubSection, EndSection, SubTokenizer)
from kaa.theme import Theme, Style
from kaa.filetype.javascript import javascriptmode
from kaa.filetype.css import cssmode

from kaa import encodingdef
from kaa.filetype import filetypedef
import re

def iter_attr(b):
    pos = 0
    prop = re.compile(br'(\w+)\s*=\s*', re.DOTALL)
    value = re.compile(b'\s*(("[^"]+")|(\'[^\']+\'))', re.DOTALL)

    while True:
        m = prop.search(b, pos)
        if not m:
            return
        name = m.group(1)
        m = value.match(b, m.end())
        if not m:
            return
        yield name, m.group()[1:-1]
        pos = m.end()

def get_encoding(b):
    # http://www.w3.org/International/questions/qa-html-encoding-declarations

    # HTML5: <meta charset="UTF-8">
    m = re.search(br'<meta ', b, re.DOTALL)
    if m:
        for name, value in iter_attr(b[m.end():]):
            if name == b'charset':
                return str(value.strip(), 'utf-8')

    # HTML4, XHTML: 
    #    <meta http-equiv="Content-type" content="text/html;charset=UTF-8">
    m = re.search(br'<meta ', b, re.DOTALL)
    if m:
        for name, value in iter_attr(b[m.end():]):
            if name == b'content':
                m = re.search(br'charset=(.*)', value, re.DOTALL)
                if m:
                    return str(m.group(1).strip(), 'utf-8')
                break

    # XHTML: <?xml version="1.0" encoding="UTF-8"?>
    m = re.search(br'<\?xml ', b, re.DOTALL)
    if m:
        for name, value in iter_attr(b[m.end():]):
            if name == b'encoding':
                return str(value.strip().strip(), 'utf-8')


HTMLThemes = {
    'basic': [
        Style('html-decl', 'Red', 'default'),
        Style('html-tag', 'Magenta', 'default'),
        Style('html-attrname', 'green', 'default'),
        Style('html-attrvalue', 'yellow', 'default'),
    ],
}


class HTMLTag(Token):

    def re_start(self):
        return r'<'

    def prepare(self, tokenizer):
        super().prepare(tokenizer)

        self.span_lt = self.assign_tokenid(tokenizer, 'html-tag')
        self.span_gt = self.assign_tokenid(tokenizer, 'html-tag')
        self.span_elemname = self.assign_tokenid(tokenizer, 'html-tag')
        self.span_attrname = self.assign_tokenid(tokenizer, 'html-attrname')
        self.span_attrvalue = self.assign_tokenid(tokenizer, 'html-attrvalue')
        self.span_elemws = self.assign_tokenid(tokenizer, 'default')

    def resume_pos(self, highlighter, tokenizer, doc, pos):
        # Returns top of current keyword
        if 0 < pos < len(doc.styles):
            p = doc.styles.rfindint([self.span_lt],
                                    0, pos, comp_ne=False)
            if p != -1:
                return p
        return 0

    def yield_jsattr(self, tokenizer, doc, match):
        f, t = match.span('ATTRVALUE')
        if match.group('Q1'):
            func = tokenizer.tokens.jsattr2
        elif match.group('Q2'):
            func = tokenizer.tokens.jsattr1
        else:
            # no quotation mark. ignore this value.
            yield match.start(), match.end(), self.span_attrvalue
            return match.end()

        yield f, f + 1, self.span_attrvalue
        js_to, tok, close = yield from func.iter_subtokenizers(None, doc, f + 1, None)

        pos = js_to
        if js_to < doc.endpos():
            yield js_to, js_to + 1, self.span_attrvalue
            pos = js_to + 1

        return pos

    def yield_cssattr(self, tokenizer, doc, match):
        f, t = match.span('ATTRVALUE')
        if match.group('Q1'):
            func = tokenizer.tokens.cssattr2
        elif match.group('Q2'):
            func = tokenizer.tokens.cssattr1
        else:
            # no quotation mark. ignore this value.
            yield match.start(), match.end(), self.span_attrvalue
            return match.end()

        yield f, f + 1, self.span_attrvalue
        pos, tok, close = yield from func.iter_subtokenizers(None, doc, f + 1, None)
        return pos

    RE_ELEMNAME = gre.compile(r'\s*[^>\s]+')
    RE_CLOSE = gre.compile(r'\s*/?>', gre.S)

    RE_ATTRNAME = gre.compile(
        r'>|(?P<ATTRNAME>[-._:a-zA-Z0-9]+)(?P<EQUAL>\s*=)?\s*')
    RE_ATTRVALUE = gre.compile(r'\s*(?P<ATTRVALUE>({}))'.format(
        '|'.join(['[-._:a-zA-Z0-9]+', '(?P<Q1>"[^"]*")', "(?P<Q2>'[^']*')"])))

    def iter_attrs(self, tokenizer, doc, pos):
        buf = doc.buf
        while True:
            m = self.RE_ATTRNAME.search(buf, pos)
            if not m:
                return pos
            if m.group() == '>':
                return pos

            # yield attribute name
            f, t = m.span('ATTRNAME')
            yield f, t, self.span_attrname
            attrname = m.group('ATTRNAME')

            # yield after attr-name
            pos = m.end()
            if t != pos:
                yield t, pos, self.span_elemws

            # has attribute value?
            if m.group('EQUAL'):
                # yield values after '='
                m = self.RE_ATTRVALUE.match(buf, pos)
                if m:
                    attrvalue = m.group('ATTRVALUE')
                    f, t = m.span('ATTRVALUE')

                    if attrname.lower().startswith('on'):
                        pos = yield from self.yield_jsattr(tokenizer, doc, m)
                    elif attrname.lower() == 'style':
                        pos = yield from self.yield_cssattr(tokenizer, doc, m)
                    else:
                        yield f, t, self.span_attrvalue
                        pos = t

    def iter_tokens(self, tokenizer, doc, pos):
        buf = doc.buf
        match = self.RE_ELEMNAME.match(buf, pos)
        if not match:
            yield pos, len(buf), self.span_elemws
            return
        yield (pos, match.end(), self.span_elemname)

        pos = match.end()
        pos = yield from self.iter_attrs(tokenizer, doc, pos)

        m = self.RE_CLOSE.match(buf, pos)
        if m:
            yield pos, m.end(), self.span_gt
            return m.end()
        else:
            if pos != len(buf):
                yield pos, len(buf), self.span_elemws
            return len(buf)

    def on_start(self, tokenizer, doc, pos, match):
        buf = doc.buf

        yield (match.start(), match.end(), self.span_lt)

        pos = match.end()
        for f, t, tokenid in self.iter_tokens(tokenizer, doc, match.end()):
            assert f <= t
            assert pos <= f
            if pos != f:
                yield (pos, f, self.span_elemws)
            yield (f, t, tokenid)
            pos = t
        return pos, None, False


def build_tokenizers():

    HTMLTOKENS = namedtuple('htmltokens', ['keywords', 'comment',
                                           'xmlpi', 'xmldef', 'endtag', 'jsstart', 'cssstart',
                                           'htmltag', 'jsattr1', 'jsattr2', 'cssattr1',
                                           'cssattr2'])

    keywords = SingleToken('html-entityrefs', 'keyword',
                           [r'&\w+;', r'&\#x[0-9a-hA-H]+;', r'&\#[0-9]+;'])

    comment = Span('html-comment', 'comment', r'<!--', r'--\s*>')
    xmlpi = Span('xmlpi', 'html-decl', r'<\?', r'\?>')
    xmldef = Span('xmldef', 'html-decl', r'<!', r'>')

    endtag = Span('html-endtag', 'html-tag', r'</', r'>')
    htmltag = HTMLTag('html', 'default')

    jsstop = EndSection('end-javascript', 'html-tag', r'</\s*script\s*>')
    jstokenizer = javascriptmode.build_tokenizer(jsstop)

    jsstart = SubSection('start-javascript', 'html-tag',
                         r'<\s*script(\s+[^>]*)*>', jstokenizer)

    csstokenizer = cssmode.build_tokenizer(r'</\s*style\s*>', 'html-tag')

    cssstart = SubSection('start-css', 'html-tag',
                          r'<\s*style(\s+[^>]*)*>', csstokenizer)

    jsattr1 = SubTokenizer(
        'sub-js', '', javascriptmode.build_tokenizer(terminates="'"))
    jsattr2 = SubTokenizer(
        'sub-js', '', javascriptmode.build_tokenizer(terminates='"'))

    styleattr1 = SubTokenizer(
        'sub-css', '', cssmode.build_proptokenizer("'", 'html-attrvalue'))
    styleattr2 = SubTokenizer(
        'sub-css', '', cssmode.build_proptokenizer('"', 'html-attrvalue'))

    return [Tokenizer(
        HTMLTOKENS(
            keywords, comment, xmlpi, xmldef, endtag, jsstart, cssstart, htmltag,
            jsattr1, jsattr2, styleattr1, styleattr2)),
            jstokenizer, csstokenizer]


class HTMLMode(defaultmode.DefaultMode):
    MODENAME = 'HTML'

    @classmethod
    def update_fileinfo(cls, fileinfo, document=None):
        try:
            buffer = open(fileinfo.fullpathname, 'rb').read(1024)
            enc = get_encoding(buffer)
            if enc:
                enc = encodingdef.normalize_encname(enc)
                fileinfo.encoding = enc
        except IOError:
            pass


    def init_themes(self):
        super().init_themes()
        self.themes.append(HTMLThemes)
        self.themes.append(cssmode.CSSThemes)
        self.themes.append(javascriptmode.JavaScriptThemes)

    def init_tokenizers(self):
        self.tokenizers = build_tokenizers()



import collections
from kaa import document
from kaa.ui.dialog import dialogmode
from kaa.keyboard import *
from kaa.command import command, norec, norerun


SelectItem = collections.namedtuple(
    'SelectItem', ['style', 'activestyle', 'text', 'value'])

selectlist_keys = {
    down: 'selectitemlist.next',
    (ctrl, 'n'): 'selectitemlist.next',
    (ctrl, 'f'): 'selectitemlist.next',
    tab: 'selectitemlist.next',
    up: 'selectitemlist.prev',
    (ctrl, 'p'): 'selectitemlist.prev',
    (ctrl, 'b'): 'selectitemlist.prev',
    (shift, tab): 'selectitemlist.prev',
}


class SelectItemList(dialogmode.DialogMode):
    USE_UNDO = False
    NO_WRAPINDENT = False
    CAPTION_STYLE = 'caption'

    items = ()
    cursel = None
    filterfunc = None
    caption = None
    SEP = ' '

    @classmethod
    def build(cls):
        buf = document.Buffer()
        doc = document.Document(buf)
        mode = cls()
        doc.setmode(mode)

        return doc

    def is_cursor_visible(self):
        return 0   # hide cursor

    def init_keybind(self):
        super().init_keybind()
        self.keybind.add_keybind(selectlist_keys)

    def on_str(self, wnd, s):
        pass

    def calc_position(self, wnd):
        height = wnd.screen.get_total_height()
        height = min(height, wnd.mainframe.height // 2)
        top = wnd.mainframe.height - height - wnd.mainframe.MESSAGEBAR_HEIGHT
        return 0, top, wnd.mainframe.width, top + height

    def update_doc(self, items):
        self.items = list(collections.OrderedDict((i, 1)
                          for i in items).keys())

        self.cursel = None

        self.document.marks.clear()
        self.document.delete(0, self.document.endpos())
        with dialogmode.FormBuilder(self.document) as f:
            if self.caption:
                f.append_text(self.CAPTION_STYLE, self.caption + ':\n')

            for n, item in enumerate(self.items):
                f.append_text(item.style, item.text, mark_pair=item)
                if n != (len(self.items) - 1):
                    f.append_text('default', self.SEP)

    def _update_item_style(
            self, wnd, item, activate, middle=None, bottom=None):

        if item not in self.document.marks:
            return

        if activate:
            style = item.activestyle
        else:
            style = item.style

        f, t = self.document.marks[item]
        self.document.setstyles(f, t, self.get_styleid(style))
        if activate:
            wnd.screen.apply_updates()
            top = not middle and not bottom
            wnd.screen.locate(f, top=top, middle=middle, bottom=bottom)
            wnd.update_window()
            
    def update_sel(self, wnd, newsel, middle=None, bottom=None):
        if self.cursel is not None:
            self._update_item_style(wnd, self.cursel, False)

        if newsel is not None:
            self._update_item_style(wnd, newsel, True,
                                    middle=middle, bottom=bottom)

        self.cursel = newsel

    @command('selectitemlist.next')
    @norec
    @norerun
    def sel_next(self, wnd):
        if not self.items:
            newsel = None
        elif self.cursel is None:
            newsel = self.items[0]
        else:
            try:
                idx = self.items.index(self.cursel)
            except ValueError:
                newsel = self.items[0]
            else:
                if idx < len(self.items) - 1:
                    newsel = self.items[idx + 1]
                else:
                    newsel = self.items[0]

        self.update_sel(wnd, newsel, bottom=True)
        return newsel

    @command('selectitemlist.prev')
    @norec
    @norerun
    def sel_prev(self, wnd):
        bottom = None
        if not self.items:
            newsel = None
        elif self.cursel is None:
            newsel = self.items[-1]
            bottom = True
        else:
            try:
                idx = self.items.index(self.cursel)
            except ValueError:
                newsel = self.items[-1]
                bottom = True
            else:
                if idx > 0:
                    newsel = self.items[idx - 1]
                else:
                    newsel = self.items[-1]
                    bottom = True

        self.update_sel(wnd, newsel, bottom=bottom)
        return newsel
